import torch
import numpy as np


from graph4nlp.pytorch.data.dataset import Text2TextDataset, TextToTreeDataset
from ..modules.graph_construction.base import GraphConstructionBase
from ..modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
from ..data.data import GraphData


dataset_root = '../test/dataset/jobs'

class JobsDataset(Text2TextDataset):
    @property
    def raw_file_names(self):
        """3 reserved keys: 'train', 'val' (optional), 'test'. Represent the split of dataset."""
        return {'train': 'train.txt', 'test': 'test.txt'}

    @property
    def processed_file_names(self):
        """At least 3 reserved keys should be fiiled: 'vocab', 'data' and 'split_ids'."""
        return {'vocab': 'vocab.pt', 'data': 'data.pt'}

    def download(self):
        # raise NotImplementedError(
        #     'This dataset is now under test and cannot be downloaded. Please prepare the raw data yourself.')
        return

    def __init__(self, root_dir,
                 topology_builder, topology_subdir,
                 pretrained_word_emb_file=None,
                 val_split_ratio=None,
                 graph_type='static',
                 merge_strategy="tailhead", edge_strategy=None,
                 seed=None,
                 word_emb_size=300, share_vocab=True,
                 dynamic_graph_type=None,
                 dynamic_init_topology_builder=None,
                 dynamic_init_topology_aux_args=None):
        """

        Parameters
        ----------
        root_dir: str
            The path of dataset.
        topology_builder: GraphConstructionBase
            The graph construction class.
        topology_subdir: str
            The directory name of processed path.
        graph_type: str, default='static'
            The graph type. Expected in ('static', 'dynamic')
        edge_strategy: str, default=None
            The edge strategy. Expected in (None, 'homogeneous', 'as_node'). If set `None`, it will be 'homogeneous'.
        merge_strategy: str, default=None
            The strategy to merge sub-graphs. Expected in (None, 'tailhead', 'user_define').
            If set `None`, it will be 'tailhead'.
        share_vocab: bool, default=False
            Whether to share the input vocabulary with the output vocabulary.
        dynamic_graph_type: str, default=None
            The dynamic graph type. It is only available when `graph_type` is set 'dynamic'.
            Expected in (None, 'node_emb', 'node_emb_refined').
        init_graph_type: str, default=None
            The initial graph topology. It is only available when `graph_type` is set 'dynamic'.
            Expected in (None, 'dependency', 'constituency')
        """
        # Initialize the dataset. If the preprocessed files are not found, then do the preprocessing and save them.
        super(JobsDataset, self).__init__(root_dir=root_dir, topology_builder=topology_builder,
                                          topology_subdir=topology_subdir, graph_type=graph_type,
                                          edge_strategy=edge_strategy, merge_strategy=merge_strategy,
                                          share_vocab=share_vocab, pretrained_word_emb_file=pretrained_word_emb_file,
                                          val_split_ratio=val_split_ratio, seed=seed, word_emb_size=word_emb_size,

                                          dynamic_graph_type=dynamic_graph_type,
                                          dynamic_init_topology_builder=dynamic_init_topology_builder,
                                          dynamic_init_topology_aux_args=dynamic_init_topology_aux_args)

    # @classmethod
    # def from_args(cls, args, graph_type, topology_builder, dynamic_graph_type=None, dynamic_init_topology_builder=None,
    #               dynamic_init_topology_aux_args=None):
    #     return cls(root_dir=args.root_dir,
    #                pretrained_word_emb_file=args.pretrained_word_emb_file,
    #                val_split_ratio=args.val_split_ratio,
    #                merge_strategy=args.merge_strategy, edge_strategy=args.edge_strategy,
    #                seed=args.seed,
    #                word_emb_size=args.word_emb_size, share_vocab=args.share_vocab,
    #                graph_type=graph_type,
    #                topology_builder=topology_builder, topology_subdir=args.topology_subdir,
    #                dynamic_graph_type= dynamic_graph_type,
    #                dynamic_init_topology_builder=dynamic_init_topology_builder,
    #                dynamic_init_topology_aux_args=dynamic_init_topology_aux_args)




class JobsDatasetForTree(TextToTreeDataset):
    @property
    def raw_file_names(self):
        """3 reserved keys: 'train', 'val' (optional), 'test'. Represent the split of dataset."""
        return {'train': 'train.txt', 'test': 'test.txt'}

    @property
    def processed_file_names(self):
        """At least 3 reserved keys should be fiiled: 'vocab', 'data' and 'split_ids'."""
        return {'vocab': 'vocab.pt', 'data': 'data.pt'}

    def download(self):
        # raise NotImplementedError(
        #     'This dataset is now under test and cannot be downloaded. Please prepare the raw data yourself.')
        return

    def __init__(self, root_dir,
                 topology_builder, topology_subdir,
                 pretrained_word_emb_file=None,
                 val_split_ratio=0,
                 graph_type='static',
                 merge_strategy="tailhead", edge_strategy=None,
                 seed=None,
                 word_emb_size=300, share_vocab=True,
                 dynamic_graph_type=None,
                 dynamic_init_topology_builder=None,
                 dynamic_init_topology_aux_args=None,
                 enc_emb_size=300,
                 dec_emb_size=300,
                 device='cpu',
                 min_freq=1):
        """

        Parameters
        ----------
        root_dir: str
            The path of dataset.
        topology_builder: GraphConstructionBase
            The graph construction class.
        topology_subdir: str
            The directory name of processed path.
        graph_type: str, default='static'
            The graph type. Expected in ('static', 'dynamic')
        edge_strategy: str, default=None
            The edge strategy. Expected in (None, 'homogeneous', 'as_node'). If set `None`, it will be 'homogeneous'.
        merge_strategy: str, default=None
            The strategy to merge sub-graphs. Expected in (None, 'tailhead', 'user_define').
            If set `None`, it will be 'tailhead'.
        share_vocab: bool, default=False
            Whether to share the input vocabulary with the output vocabulary.
        dynamic_graph_type: str, default=None
            The dynamic graph type. It is only available when `graph_type` is set 'dynamic'.
            Expected in (None, 'node_emb', 'node_emb_refined').
        init_graph_type: str, default=None
            The initial graph topology. It is only available when `graph_type` is set 'dynamic'.
            Expected in (None, 'dependency', 'constituency')
        """
        # Initialize the dataset. If the preprocessed files are not found, then do the preprocessing and save them.
        super(JobsDatasetForTree, self).__init__(root_dir=root_dir, topology_builder=topology_builder,
                                          topology_subdir=topology_subdir, graph_type=graph_type,
                                          edge_strategy=edge_strategy, merge_strategy=merge_strategy,
                                          share_vocab=share_vocab, pretrained_word_emb_file=pretrained_word_emb_file,
                                          val_split_ratio=val_split_ratio, seed=seed, word_emb_size=word_emb_size,

                                          dynamic_graph_type=dynamic_graph_type,
                                          dynamic_init_topology_builder=dynamic_init_topology_builder,
                                          dynamic_init_topology_aux_args=dynamic_init_topology_aux_args,
                                          enc_emb_size=enc_emb_size, dec_emb_size=dec_emb_size, device=device,
                                          min_freq=min_freq)


if __name__ == '__main__':
    jobs_dataset = JobsDataset(root_dir='../test/dataset/jobs', topology_builder=DependencyBasedGraphConstruction,
                               topology_subdir='DependencyGraph')
    # Load train, val and test subsets
    train_size = len(jobs_dataset.split_ids['train'])
    # Since the validation file is not present in this example, we use the 80% of the original training set as the
    # real training set and the rest 20% as the validation set
    train_set = jobs_dataset[jobs_dataset.split_ids['train'][:int(0.8 * train_size)]]
    val_set = jobs_dataset[jobs_dataset.split_ids['train'][int(0.8 * train_size):]]

    test_set = jobs_dataset[jobs_dataset.split_ids['test']]

    from torch.utils.data import dataloader

    train_dataloader = dataloader.DataLoader(dataset=train_set, batch_size=10, shuffle=True,
                                             collate_fn=jobs_dataset.collate_fn)
    print('The number of batches in train_dataloader is {} with batch size of 10.'.format(len(train_dataloader)))

    # You can also use the built-in shuffle() method to obtain a shuffled


import math
import torch
import sympy
import random
import graph_utils

import numpy as np
import pickle as pkl

from random import randint
from operator import itemgetter
from sympy.parsing.sympy_parser import parse_expr

class SymbolsManager():
    def __init__(self, whether_add_special_tags):
        self.symbol2idx = {}
        self.idx2symbol = {}
        self.vocab_size = 0
        self.whether_add_special_tags = whether_add_special_tags
        if whether_add_special_tags:
            # PAD: padding token = 0
            self.add_symbol('<P>')
            # GO: start token = 1
            self.add_symbol('<S>')
            # EOS: end token = 2
            self.add_symbol('<E>')
            # UNK: unknown token = 3
            self.add_symbol('<U>')
            # NON: non-terminal token = 4
            self.add_symbol('<N>')

    def add_symbol(self,s):
        if s not in self.symbol2idx:
            self.symbol2idx[s] = self.vocab_size
            self.idx2symbol[self.vocab_size] = s
            self.vocab_size += 1
        return self.symbol2idx[s]

    def get_symbol_idx(self,s):
        if s not in self.symbol2idx:
            if self.whether_add_special_tags:
                return self.symbol2idx['<U>']
            else:
                print("not reached!")
                return 0
        return self.symbol2idx[s]

    def get_idx_symbol(self, idx):
        if idx not in self.idx2symbol:
            return '<U>'
        return self.idx2symbol[idx]

    def init_from_file(self, fn, min_freq, max_vocab_size):
        # the vocab file is sorted by word_freq
        print("loading vocabulary file: {}".format(fn))
        with open(fn,"r") as f:
            for line in f:
                l_list = line.strip().split('\t')
                c = int(l_list[1])
                if c >= min_freq:
                    self.add_symbol(l_list[0])
                if self.vocab_size > max_vocab_size:
                    break

    def get_symbol_idx_for_list(self,l):
        r = []
        for i in range(len(l)):
            r.append(self.get_symbol_idx(l[i]))
        return r

class MinibatchLoader_for_seq2seq():
    def __init__(self, opt, mode, using_gpu, word_manager):
        if opt.pointer_generator:
            data = pkl.load(open("{}/share_{}.pkl".format(opt.data_dir, mode), "rb" ))
        else:
            data = pkl.load(open("{}/{}.pkl".format(opt.data_dir, mode), "rb" ))
        if len(data) % opt.batch_size != 0:
            n = len(data)
            for i in range(len(data)%opt.batch_size):
                data.insert(n-i-1, data[n-i-1])
        graph_data = graph_utils.read_graph_data("{}/graph.{}".format(opt.data_dir, mode))
        if len(graph_data) % opt.batch_size != 0:
            n = len(graph_data)
            for i in range(len(graph_data)%opt.batch_size):
                graph_data.insert(n-i-1, graph_data[n-i-1])

        self.enc_batch_list = []
        self.enc_len_batch_list = []
        self.dec_batch_list = []
        p = 0
        while p + opt.batch_size <= len(data):
            # build encoder matrix

            batch_graph = [graph_data[p + idx] for idx in range(opt.batch_size)]
            combine_batch_graph = graph_utils.cons_batch_graph(batch_graph)
            vector_batch_graph = graph_utils.vectorize_batch_graph(combine_batch_graph,word_manager)

            self.enc_batch_list.append(vector_batch_graph)
            # print(vector_batch_graph['g_fw_adj'].shape)
            self.enc_len_batch_list.append(len(batch_graph[0]['g_ids']))
            # build decoder matrix
            max_len = -1
            for i in range(opt.batch_size):
                w_list = data[p+i][1]
                if len(w_list) > max_len:
                    max_len = len(w_list)
            m_text = torch.zeros((opt.batch_size, max_len + 2), dtype=torch.long)
            if using_gpu:
                m_text = m_text.cuda()
            # add <S>
            m_text[:,0] = 1
            for i in range(opt.batch_size):
                w_list = data[p+i][1]
                for j in range(len(w_list)):
                    m_text[i][j+1] = w_list[j]
                # add <E>
                m_text[i][len(w_list)+1] = 2
            self.dec_batch_list.append(m_text)
            p += opt.batch_size
        self.num_batch = len(self.enc_batch_list)
        assert(len(self.enc_batch_list) == len(self.dec_batch_list))

    def random_batch(self):
        p = randint(0,self.num_batch-1)
        return self.enc_batch_list[p], self.enc_len_batch_list[p], self.dec_batch_list[p]

    def get_batch0(self):
        p = 0
        return self.enc_batch_list[p], self.enc_len_batch_list[p], self.dec_batch_list[p]

    def all_batch(self):
        r = []
        for p in range(self.num_batch):
            r.append([self.enc_batch_list[p], self.enc_len_batch_list[p], self.dec_batch_list[p]])
        return r

def is_all_same(c1, c2, form_manager):
    all_same = False
    if len(c1) == len(c2):
        all_same = True
        for j in range(len(c1)):
            if c1[j] != c2[j]:
                all_same = False
                break
    if all_same == False:
        if is_solution_same(c1, c2, form_manager):
            return True
        return False
    else:
        return True

def is_solution_same(i1, i2, form_manager):
    c1 = " ".join([form_manager.get_idx_symbol(x) for x in i1])
    c2 = " ".join([form_manager.get_idx_symbol(x) for x in i2])
    if ('=' not in c1) or ('=' not in c2):
        return False
    elif ('<U>' in c1) or ('<U>' in c2):
        return False
    else:
        try:
            s1 = c1.split('=')
            s2 = c2.split('=')
            eq1 = []
            eq2 = []
            x = sympy.Symbol('x')
            eq1.append(parse_expr(s1[0]))
            eq1.append(parse_expr(s1[1]))
            eq2.append(parse_expr(s2[0]))
            eq2.append(parse_expr(s2[1]))
            res1 = sympy.solve(sympy.Eq(eq1[0], eq1[1]), x)
            res2 = sympy.solve(sympy.Eq(eq2[0], eq2[1]), x)
            if not res1 or not res2:
                return False
            return res1[0] == res2[0]

        except BaseException:
            print(c1)
            print(c2)
            return False

def compute_accuracy(candidate_list, reference_list, form_manager):
    if len(candidate_list) != len(reference_list):
        print("candidate list has length {}, reference list has length {}\n".format(len(candidate_list), len(reference_list)))

    len_min = min(len(candidate_list), len(reference_list))
    c = 0
    for i in range(len_min):
        if is_all_same(candidate_list[i], reference_list[i], form_manager):
            c = c+1
        else:
            pass
    return c/float(len_min)



import os
import pickle
import re
from collections import Counter
from functools import lru_cache

import numpy as np
from nltk.tokenize import word_tokenize

from . import constants

word_detector = re.compile('\w')


class VocabModel(object):
    """Vocab model builder.

    Parameters
    ----------
    data_set: iterable
        A list of instances where each instance is a list of str.
    tokenizer: function, optional
        Word tokenization function, default: nltk.tokenize.word_tokenize.
    max_word_vocab_size: int, optional
        Maximal word vocab size, default: ``None``.
    min_word_vocab_freq: int, optional
        Minimal word vocab frequency, default: ``1``.
    pretrained_word_emb_file: str, optional
        Path to the pretrained word embedding file, default: ``None``.
    word_emb_size: int, optional
        Word embedding size, default: ``None``.

    Examples
    -------
    # Build a vocab model from scratch
    >>> vocab_model = VocabModel([['I like nlp.', 'Same here!'],
                        ['I like graph.', 'Same here!']],
                        max_word_vocab_size=None,
                        min_word_vocab_freq=1,
                        word_emb_size=300)
    >>> print(vocab_model.word_vocab.get_vocab_size())

    # Restore a vocab model from disk if exists or build one from scratch.
    >>> vocab_model = VocabModel.build('vocab_model.pkl', [['I like nlp.', 'Same here!'],
                            ['I like graph.', 'Same here!']],
                            max_word_vocab_size=None,
                            min_word_vocab_freq=1,
                            word_emb_size=300)
    >>> print(vocab_model.word_vocab.get_vocab_size())
    """

    def __init__(self, data_set=None,
                 tokenizer=word_tokenize,
                 lower_case=True,
                 max_word_vocab_size=None,
                 min_word_vocab_freq=1,
                 pretrained_word_emb_file=None,
                 word_emb_size=None,
                 share_vocab=True):
        super(VocabModel, self).__init__()
        self.tokenizer = tokenizer

        print('Building vocabs...')
        all_words = VocabModel.collect_vocabs(data_set, self.tokenizer, lower_case=lower_case, share_vocab=share_vocab)
        # print('Number of words: {}'.format(len(all_words)))
        if share_vocab:
            in_all_words, out_all_words = all_words, None
        else:
            in_all_words, out_all_words = all_words


        self.in_word_vocab = Vocab(lower_case=lower_case, tokenizer=self.tokenizer)
        self.in_word_vocab.build_vocab(in_all_words, max_vocab_size=max_word_vocab_size, min_vocab_freq=min_word_vocab_freq)

        if pretrained_word_emb_file is not None:
            self.in_word_vocab.load_embeddings(pretrained_word_emb_file)
            print('Using pretrained word embeddings')
        else:
            self.in_word_vocab.randomize_embeddings(word_emb_size)

        if out_all_words is not None:
            self.out_word_vocab = Vocab(lower_case=lower_case, tokenizer=self.tokenizer)
            self.out_word_vocab.build_vocab(out_all_words, max_vocab_size=max_word_vocab_size, min_vocab_freq=min_word_vocab_freq)

            if pretrained_word_emb_file is not None:
                self.out_word_vocab.load_embeddings(pretrained_word_emb_file)
                print('Using pretrained word embeddings')
            else:
                self.out_word_vocab.randomize_embeddings(word_emb_size)
        else:
            self.out_word_vocab = self.in_word_vocab

        if share_vocab:
            print('[ Initialized word embeddings: {} ]'.format(self.in_word_vocab.embeddings.shape))
        else:
            print('[ Using separate word vocabs for input & output text ]')
            print('[ Initialized input word embeddings: {} ]'.format(self.in_word_vocab.embeddings.shape))
            print('[ Initialized output word embeddings: {} ]'.format(self.out_word_vocab.embeddings.shape))

        # self.edge_vocab = Vocab()
        # self.edge_vocab.build_vocab(all_edge_types)
        # print('edge_vocab: {}'.format((self.edge_vocab.get_vocab_size())))

        # self.POS_vocab = Vocab()
        # self.POS_vocab.build_vocab(all_POSs)
        # print('POS_vocab: {}'.format(self.POS_vocab.get_vocab_size()))

        # self.NER_vocab = Vocab()
        # self.NER_vocab.build_vocab(all_NERs)
        # print('NER_vocab: {}'.format(self.NER_vocab.get_vocab_size()))

    @classmethod
    def build(cls, saved_vocab_file,
              data_set=None,
              tokenizer=word_tokenize,
              lower_case=True,
              max_word_vocab_size=None,
              min_word_vocab_freq=1,
              pretrained_word_emb_file=None,
              word_emb_size=None,
              share_vocab=True):
        """Static method for loading a VocabModel from disk.

        Parameters:
        -------
        saved_vocab_file : str
            Path to the saved vocab file.
        data_set: iterable
            A list of instances where each instance is a list of str.
        tokenizer: function, optional
            Word tokenization function, default: nltk.tokenize.word_tokenize.
        max_word_vocab_size: int, optional
            Maximal word vocab size, default: ``None``.
        min_word_vocab_freq: int, optional
            Minimal word vocab frequency, default: ``1``.
        pretrained_word_emb_file: str, optional
            Path to the pretrained word embedding file, default: ``None``.
        word_emb_size: int, optional
            Word embedding size, default: ``None``.

        Returns:
        -------
        VocabModel
            Loaded Vocabulary.
        """
        if os.path.exists(saved_vocab_file):
            print('Loading pre-built vocab model stored in {}'.format(saved_vocab_file))
            with open(saved_vocab_file, 'rb') as f:
                vocab_model = pickle.load(f)

        else:
            vocab_model = VocabModel(data_set=data_set, tokenizer=tokenizer,
                                     max_word_vocab_size=max_word_vocab_size,
                                     min_word_vocab_freq=min_word_vocab_freq,
                                     pretrained_word_emb_file=pretrained_word_emb_file,
                                     word_emb_size=word_emb_size,
                                     share_vocab=share_vocab)
            print('Saving vocab model to {}'.format(saved_vocab_file))
            pickle.dump(vocab_model, open(saved_vocab_file, 'wb'))

        return vocab_model

    @staticmethod
    def collect_vocabs(all_instances, tokenizer, lower_case=True, share_vocab=True):
        """Count vocabulary tokens."""
        if share_vocab:
            all_words = Counter()
        else:
            all_words = [Counter(), Counter()]

        for instance in all_instances:
            extracted_tokens = instance.extract()
            if share_vocab:
                all_words.update(extracted_tokens)
            else:
                all_words[0].update(extracted_tokens[0])
                all_words[1].update(extracted_tokens[1])

        return all_words


class Vocab(object):
    """Vocab class.

    Parameters
    ----------
    tokenizer: function, optional
        Word tokenization function, default: nltk.tokenize.word_tokenize.

    Examples
    -------
    >>> word_vocab = Vocab()
    >>> word_vocab.build_vocab({'i': 10, 'like': 5, 'nlp': 3})
    >>> print(word_vocab.get_vocab_size())
    """
    PAD = 0
    SOS = 1
    EOS = 2
    UNK = 3
    pad_token = constants._PAD_TOKEN
    sos_token = constants._SOS_TOKEN
    eos_token = constants._EOS_TOKEN
    unk_token = constants._UNK_TOKEN
    def __init__(self, lower_case=True, tokenizer=word_tokenize):
        super(Vocab, self).__init__()
        self.lower_case = lower_case
        self.tokenizer = tokenizer
        self.reserved = [self.pad_token, self.sos_token, self.eos_token, self.unk_token]
        self.index2word = self.reserved[:]
        self.word2index = dict(zip(self.reserved, range(len(self.reserved))))
        self.word2count = Counter()
        self.embeddings = None

    def build_vocab(self, vocab_counter, max_vocab_size=None, min_vocab_freq=1):
        """Build vocab from ``vocab_counter`` which is a vocab count dict.

        Parameters
        ----------
        vocab_counter : dict
            Vocab counter.
        max_vocab_size : None, optional
            Maximal word vocab size, default: ``None``.
        min_vocab_freq : int, optional
            Minimal word vocab frequency, default: ``1``.
        """
        self.word2count = vocab_counter
        self._add_words(vocab_counter.keys())
        self._trim(max_vocab_size=max_vocab_size, min_vocab_freq=min_vocab_freq)

    def _add_words(self, words):
        # words: a list of str
        for word in words:
            if self.lower_case:
                word = word.lower()

            if word not in self.word2index:
                self.word2index[word] = len(self.index2word)
                self.index2word.append(word)
        assert len(self.word2index) == len(self.index2word)

    def _trim(self, max_vocab_size=None, min_vocab_freq=1):
        """Trim vocab"""
        if min_vocab_freq <= 1 and (max_vocab_size is None or max_vocab_size >= len(self.word2index)):
            return

        ordered_words = sorted(((c, w) for (w, c) in self.word2count.items()), reverse=True)

        if max_vocab_size:
            ordered_words = ordered_words[:max_vocab_size]

        self.index2word = self.reserved[:]
        self.word2index = dict(zip(self.reserved, range(len(self.reserved))))
        self.word2count = Counter()

        for count, word in ordered_words:
            if count < min_vocab_freq: break
            if word not in self.word2index:
                self.word2index[word] = len(self.index2word)
                self.word2count[word] = count
                self.index2word.append(word)

        assert len(self.word2index) == len(self.index2word)

    def load_embeddings(self, file_path, scale=0.08, dtype=np.float32):
        """Load pretrained word embeddings for initialization"""
        hit_words = set()
        vocab_size = len(self)
        with open(file_path, 'rb') as f:
            for line in f:
                line = line.split()
                word = line[0].decode('utf-8')
                if self.lower_case:
                    word = word.lower()

                idx = self.word2index.get(word, None)
                if idx is None or idx in hit_words:
                    continue

                vec = np.array(line[1:], dtype=dtype)
                if self.embeddings is None:
                    n_dims = len(vec)
                    self.embeddings = np.array(np.random.uniform(low=-scale, high=scale, size=(vocab_size, n_dims)),
                                               dtype=dtype)
                    self.embeddings[self.PAD] = np.zeros(n_dims)
                self.embeddings[idx] = vec
                hit_words.add(idx)
        print('Pretrained word embeddings hit ratio: {}'.format(len(hit_words) / len(self.index2word)))

    def randomize_embeddings(self, n_dims, scale=0.08):
        """Use random word embeddings for initialization."""
        vocab_size = self.get_vocab_size()
        shape = (vocab_size, n_dims)
        self.embeddings = np.array(np.random.uniform(low=-scale, high=scale, size=shape), dtype=np.float32)
        self.embeddings[self.PAD] = np.zeros(n_dims)

    def __getitem__(self, item):
        if type(item) is int:
            return self.index2word[item]
        return self.word2index.get(item, self.UNK)

    def __len__(self):
        return len(self.index2word)

    @lru_cache(maxsize=None)
    def is_word(self, token_id: int) -> bool:
        """Return whether the token at `token_id` is a word; False for punctuations."""
        if token_id < 4: return False
        if token_id >= len(self): return True  # OOV is assumed to be words
        token_str = self.index2word[token_id]
        if not word_detector.search(token_str) or token_str == '<P>':
            return False
        return True

    def get_vocab_size(self):
        return len(self.index2word)

    def getIndex(self, word, use_ie=False):
        # For IE Graph
        # word can be a phrase
        if use_ie:
            if self.lower_case:
                word = word.lower()

            ret = []
            for x in word.replace('_',' ').split(' '):
                if x == '':
                    continue
                ret.append(self.word2index.get(x, self.UNK))

            return ret

        if self.lower_case:
            word = word.lower()

        return self.word2index.get(word, self.UNK)

    def getWord(self, idx):
        return self.index2word[idx] if idx < len(self.index2word) else self.unk_token

    def to_word_sequence(self, seq):
        sentence = []
        for idx in seq:
            word = self.getWord(idx)
            sentence.append(word)
        return ' '.join(sentence)

    def to_index_sequence(self, sentence):
        sentence = sentence.strip()
        if self.lower_case:
            sentence = sentence.lower()

        seq = []
        if self.tokenizer is None:
            for word in sentence.split():
                idx = self.getIndex(word)
                seq.append(idx)
        else:
            for word in self.tokenizer(sentence):
                idx = self.getIndex(word)
                seq.append(idx)
        return seq

    def to_index_sequence_for_list(self, words):
        seq = []
        for word in words:
            if self.lower_case:
                word = word.lower()

            idx = self.getIndex(word)
            seq.append(idx)
        return seq



from .base import EvaluationMetricBase
import sacrebleu


class BLEUTranslation(EvaluationMetricBase):
    def __init__(self):
        super(BLEUTranslation, self).__init__()

    def calculate_scores(self, ground_truth, predict):
        """
            The standard BLEU calculation function for translation. It will compute the BLEU scores using sacrebleu tools.
        Parameters
        ----------
        ground_truth: list[string]
            The ground truth (correct) target values. It is a list of strings.
        predict: list[string]
            The predicted target values. It is a list of strings.
        Returns
        -------
        score: float
            The final bleu score
        """
        bleu = sacrebleu.corpus_bleu(predict, [ground_truth], lowercase=True)
        return bleu.score



import numpy as np
from scipy import sparse
import torch
from torch import nn
import torch.nn.functional as F

from .embedding_construction import EmbeddingConstruction
from ..utils.constants import INF
from ..utils.generic_utils import normalize_sparse_adj, sparse_mx_to_torch_sparse_tensor, to_cuda
from ..utils.constants import VERY_SMALL_NUMBER


class GraphConstructionBase(nn.Module):
    """Base class for graph construction.

    Parameters
    ----------
    word_vocab : Vocab
        The word vocabulary.
    embedding_styles : dict
        - ``single_token_item`` : specify whether the item (i.e., node or edge) contains single token or multiple tokens.
        - ``emb_strategy`` : specify the embedding construction strategy.
        - ``num_rnn_layers``: specify the number of RNN layers.
        - ``bert_model_name``: specify the BERT model name.
        - ``bert_lower_case``: specify whether to lower case the input text for BERT embeddings.
    hidden_size : int, optional
        The hidden size of RNN layer, default: ``None``.
    fix_word_emb : boolean, optional
        Specify whether to fix pretrained word embeddings, default: ``True``.
    fix_bert_emb : boolean, optional
        Specify whether to fix pretrained BERT embeddings, default: ``True``.
    bert_model_name : str, optional
        Specify the BERT model name, default: ``'bert-base-uncased'``.
    bert_lower_case : bool, optional
        Specify whether to lower case the input text for BERT embeddings, default: ``True``.
    word_dropout : float, optional
        Dropout ratio for word embedding, default: ``None``.
    rnn_dropout : float, optional
        Dropout ratio for RNN embedding, default: ``None``.
    device : torch.device, optional
        Specify computation device (e.g., CPU), default: ``None`` for using CPU.
    """
    def __init__(self,
                word_vocab,
                embedding_styles,
                hidden_size=None,
                fix_word_emb=True,
                fix_bert_emb=True,
                word_dropout=None,
                rnn_dropout=None,
                device=None):
        super(GraphConstructionBase, self).__init__()
        self.embedding_layer = EmbeddingConstruction(word_vocab,
                                        embedding_styles['single_token_item'],
                                        emb_strategy=embedding_styles['emb_strategy'],
                                        hidden_size=hidden_size,
                                        num_rnn_layers=embedding_styles.get('num_rnn_layers', 1),
                                        fix_word_emb=fix_word_emb,
                                        fix_bert_emb=fix_bert_emb,
                                        bert_model_name=embedding_styles.get('bert_model_name', 'bert-base-uncased'),
                                        bert_lower_case=embedding_styles.get('bert_lower_case', True),
                                        word_dropout=word_dropout,
                                        rnn_dropout=rnn_dropout,
                                        device=device)

    def forward(self, raw_text_data, **kwargs):
        """Compute graph topology and initial node/edge embeddings.

        Parameters
        ----------
        raw_text_data :
            The raw text data.
        **kwargs
            Extra parameters.

        Raises
        ------
        NotImplementedError
            NotImplementedError.
        """
        raise NotImplementedError()

    def topology(self, **kwargs):
        """Compute graph topology.

        Parameters
        ----------
        **kwargs
            Extra parameters.

        Raises
        ------
        NotImplementedError
            NotImplementedError.
        """
        raise NotImplementedError()

    def embedding(self, **kwargs):
        """Compute initial node/edge embeddings.

        Parameters
        ----------
        **kwargs
            Extra parameters.

        Raises
        ------
        NotImplementedError
            NotImplementedError.
        """
        raise NotImplementedError()

class StaticGraphConstructionBase(GraphConstructionBase):
    """
    Base class for static graph construction.

    ...

    Attributes
    ----------
    embedding_styles : dict
        Specify embedding styles including ``word_emb_type``, ``node_edge_emb_strategy`` and ``seq_info_encode_strategy``.

    Methods
    -------
    add_vocab()
        Add new parsed words or syntactic components into vocab.

    topology()
        Generate graph topology.

    embedding(raw_data, structure)
        Generate graph embeddings.

    forward(raw_data)
        Generate static graph embeddings and topology.
    """

    def __init__(self, word_vocab, embedding_styles, hidden_size,
                 fix_word_emb=True, fix_bert_emb=True, word_dropout=None, rnn_dropout=None, device=None):
        super(StaticGraphConstructionBase, self).__init__(word_vocab,
                                                           embedding_styles,
                                                           hidden_size,
                                                           fix_word_emb=fix_word_emb,
                                                           fix_bert_emb=fix_bert_emb,
                                                           word_dropout=word_dropout,
                                                           rnn_dropout=rnn_dropout,
                                                           device=device)

    def add_vocab(self, **kwargs):
        raise NotImplementedError()

    @classmethod
    def parsing(cls, **kwargs):
        raise NotImplementedError()

    @classmethod
    def topology(cls, **kwargs):
        raise NotImplementedError()

    def embedding(self, **kwargs):
        raise NotImplementedError()

    def forward(self, **kwargs):
        raise NotImplementedError()

    @classmethod
    def _construct_static_graph(cls, **kwargs):
        raise NotImplementedError()

    @classmethod
    def _graph_connect(cls, **kwargs):
        raise NotImplementedError()

class DynamicGraphConstructionBase(GraphConstructionBase):
    """Base class for dynamic graph construction.

    Parameters
    ----------
    word_vocab : Vocab
        The word vocabulary.
    embedding_styles : dict
        - ``single_token_item`` : specify whether the item (i.e., node or edge) contains single token or multiple tokens.
        - ``emb_strategy`` : specify the embedding construction strategy.
        - ``num_rnn_layers``: specify the number of RNN layers.
        - ``bert_model_name``: specify the BERT model name.
        - ``bert_lower_case``: specify whether to lower case the input text for BERT embeddings.
    sim_metric_type : str, optional
        Specify similarity metric function type including "attention",
        "weighted_cosine", "gat_attention", "rbf_kernel", and "cosine".
        Default: ``"weighted_cosine"``.
    num_heads : int, optional
        Specify the number of heads for multi-head similarity metric
        function, default: ``1``.
    top_k_neigh : int, optional
        Specify the top k value for knn neighborhood graph sparsificaiton,
        default: ``None``.
    epsilon_neigh : float, optional
        Specify the epsilon value (i.e., between ``0`` and ``1``) for
        epsilon neighborhood graph sparsificaiton, default: ``None``.
    smoothness_ratio : float, optional
        Specify the smoothness ratio (i.e., between ``0`` and ``1``)
        for graph regularization on smoothness, default: ``None``.
    connectivity_ratio : float, optional
        Specify the connectivity ratio (i.e., between ``0`` and ``1``)
        for graph regularization on connectivity, default: ``None``.
    sparsity_ratio : float, optional
        Specify the sparsity ratio (i.e., between ``0`` and ``1``)
        for graph regularization on sparsity, default: ``None``.
    input_size : int, optional
        The dimension of input embeddings, default: ``None``.
    hidden_size : int, optional
        The dimension of hidden layers, default: ``None``.
    fix_word_emb : boolean, optional
        Specify whether to fix pretrained word embeddings, default: ``False``.
    fix_bert_emb : boolean, optional
        Specify whether to fix pretrained BERT embeddings, default: ``True``.
    bert_model_name : str, optional
        Specify the BERT model name, default: ``'bert-base-uncased'``.
    bert_lower_case : bool, optional
        Specify whether to lower case the input text for BERT embeddings, default: ``True``.
    word_dropout : float, optional
        Dropout ratio for word embedding, default: ``None``.
    rnn_dropout : float, optional
        Dropout ratio for RNN embedding, default: ``None``.
    device : torch.device, optional
        Specify computation device (e.g., CPU), default: ``None`` for using CPU.
    """
    def __init__(self,
                word_vocab,
                embedding_styles,
                sim_metric_type='weighted_cosine',
                num_heads=1,
                top_k_neigh=None,
                epsilon_neigh=None,
                smoothness_ratio=None,
                connectivity_ratio=None,
                sparsity_ratio=None,
                input_size=None,
                hidden_size=None,
                fix_word_emb=False,
                fix_bert_emb=False,
                word_dropout=None,
                rnn_dropout=None,
                device=None):
        super(DynamicGraphConstructionBase, self).__init__(
                                                    word_vocab,
                                                    embedding_styles,
                                                    hidden_size=hidden_size,
                                                    fix_word_emb=fix_word_emb,
                                                    fix_bert_emb=fix_bert_emb,
                                                    word_dropout=word_dropout,
                                                    rnn_dropout=rnn_dropout,
                                                    device=device)
        assert top_k_neigh is None or epsilon_neigh is None, \
            'top_k_neigh and epsilon_neigh cannot be activated at the same time!'
        self.device = device
        self.top_k_neigh = top_k_neigh
        self.epsilon_neigh = epsilon_neigh
        self.sim_metric_type = sim_metric_type
        self.smoothness_ratio = smoothness_ratio
        self.connectivity_ratio = connectivity_ratio
        self.sparsity_ratio = sparsity_ratio

        if self.sim_metric_type == 'attention':
            self.mask_off_val = -INF
            self.linear_sims = nn.ModuleList([nn.Linear(input_size, hidden_size, bias=False)
                                                    for _ in range(num_heads)])
        elif self.sim_metric_type == 'weighted_cosine':
            self.mask_off_val = 0
            self.weight = torch.Tensor(num_heads, input_size)
            self.weight = nn.Parameter(nn.init.xavier_uniform_(self.weight))
        elif self.sim_metric_type == 'gat_attention':
            self.mask_off_val = -INF
            self.linear_sims1 = nn.ModuleList([nn.Linear(input_size, 1, bias=False)
                                                    for _ in range(num_heads)])
            self.linear_sims2 = nn.ModuleList([nn.Linear(input_size, 1, bias=False)
                                                    for _ in range(num_heads)])
            self.leakyrelu = nn.LeakyReLU(0.2)
        elif self.sim_metric_type == 'rbf_kernel':
            self.mask_off_val = 0
            self.precision_inv_dis = nn.Parameter(torch.Tensor(1, 1))
            self.precision_inv_dis.data.uniform_(0, 1.0)
            self.weight = nn.Parameter(nn.init.xavier_uniform_(torch.Tensor(input_size, hidden_size)))
        elif self.sim_metric_type == 'cosine':
            self.mask_off_val = 0
        else:
            raise RuntimeError('Unknown sim_metric_type: {}'.format(self.sim_metric_type))


    def forward(self, raw_text_data, **kwargs):
        """Compute graph topology and initial node/edge embeddings.

        Parameters
        ----------
        raw_text_data : list of sequences.
            The raw text data.
        **kwargs
            Extra parameters.

        Raises
        ------
        NotImplementedError
            NotImplementedError.
        """
        raise NotImplementedError()

    def topology(self, node_emb, edge_emb=None,
                    init_adj=None, node_mask=None, **kwargs):
        """Compute graph topology.

        Parameters
        ----------
        node_emb : torch.Tensor
            The node embeddings.
        edge_emb : torch.Tensor, optional
            The edge embeddings, default: ``None``.
        init_adj : torch.Tensor, optional
            The initial adjacency matrix, default: ``None``.
        node_mask : torch.Tensor, optional
            The node mask matrix, default: ``None``.
        **kwargs
            Extra parameters.

        Raises
        ------
        NotImplementedError
            NotImplementedError.
        """
        raise NotImplementedError()

    def embedding(self, feat, **kwargs):
        """Compute initial node/edge embeddings.

        Parameters
        ----------
        **kwargs
            Extra parameters.

        Raises
        ------
        NotImplementedError
            NotImplementedError.
        """
        raise NotImplementedError()

    def compute_similarity_metric(self, node_emb, node_mask=None):
        """Compute similarity metric.

        Parameters
        ----------
        node_emb : torch.Tensor
            The input node embedding matrix.
        node_mask : torch.Tensor, optional
            The node mask matrix, default: ``None``.

        Returns
        -------
        torch.Tensor
            Adjacency matrix.
        """
        if self.sim_metric_type == 'attention':
            attention = 0
            for _ in range(len(self.linear_sims)):
                node_vec_t = torch.relu(self.linear_sims[_](node_emb))
                attention += torch.matmul(node_vec_t, node_vec_t.transpose(-1, -2))

            attention /= len(self.linear_sims)
        elif self.sim_metric_type == 'weighted_cosine':
            expand_weight_tensor = self.weight.unsqueeze(1)
            if len(node_emb.shape) == 3:
                expand_weight_tensor = expand_weight_tensor.unsqueeze(1)

            node_vec_t = node_emb.unsqueeze(0) * expand_weight_tensor
            node_vec_norm = F.normalize(node_vec_t, p=2, dim=-1)
            attention = torch.matmul(node_vec_norm, node_vec_norm.transpose(-1, -2)).mean(0)
        elif self.sim_metric_type == 'gat_attention':
            attention = []
            for _ in range(len(self.linear_sims1)):
                a_input1 = self.linear_sims1[_](node_emb)
                a_input2 = self.linear_sims2[_](node_emb)
                attention.append(self.leakyrelu(a_input1 + a_input2.transpose(-1, -2)))

            attention = torch.mean(torch.stack(attention, 0), 0)
        elif self.sim_metric_type == 'rbf_kernel':
            dist_weight = torch.mm(self.weight, self.weight.transpose(-1, -2))
            attention = self._compute_distance_matrix(node_emb, dist_weight)
            attention = torch.exp(-0.5 * attention * (self.precision_inv_dis**2))
        elif self.sim_metric_type == 'cosine':
            node_vec_norm = node_emb.div(torch.norm(node_emb, p=2, dim=-1, keepdim=True))
            attention = torch.mm(node_vec_norm, node_vec_norm.transpose(-1, -2)).detach()

        if node_mask is not None:
            if torch.__version__ < '1.3.0':
                attention = attention.masked_fill_(~(node_mask == 1.), self.mask_off_val)
            else:
                attention = attention.masked_fill_(~node_mask.bool(), self.mask_off_val)


        return attention

    def sparsify_graph(self, adj):
        if self.epsilon_neigh is not None:
            adj = self._build_epsilon_neighbourhood(adj, self.epsilon_neigh)

        if self.top_k_neigh is not None:
            adj = self._build_knn_neighbourhood(adj, self.top_k_neigh)

        return adj

    def compute_graph_regularization(self, adj, node_feat):
        """Graph graph regularization loss.

        Parameters
        ----------
        adj : torch.Tensor
            The adjacency matrix.
        node_feat : torch.Tensor
            The node feature matrix.

        Returns
        -------
        torch.float32
            The graph regularization loss.
        """
        graph_reg = 0
        if not self.smoothness_ratio in (0, None):
            L = torch.diagflat(torch.sum(adj, -1)) - adj
            graph_reg += self.smoothness_ratio / int(np.prod(adj.shape))\
                    * torch.trace(torch.mm(node_feat.transpose(-1, -2), torch.mm(L, node_feat)))

        if not self.connectivity_ratio in (0, None):
            ones_vec = to_cuda(torch.ones(adj.size(-1)), self.device)
            graph_reg += -self.connectivity_ratio / adj.shape[-1]\
                    * torch.mm(ones_vec.unsqueeze(0), torch.log(torch.mm(adj, ones_vec.unsqueeze(-1)) + VERY_SMALL_NUMBER)).squeeze()

        if not self.sparsity_ratio in (0, None):
            graph_reg += self.sparsity_ratio / int(np.prod(adj.shape))\
                    * torch.sum(torch.pow(adj, 2))

        return graph_reg

    def _build_knn_neighbourhood(self, attention, top_k_neigh):
        """Build kNN neighborhood graph.

        Parameters
        ----------
        attention : torch.Tensor
            The attention matrix.
        top_k_neigh : int
            The top k value for kNN neighborhood graph.

        Returns
        -------
        torch.Tensor
            The sparsified adjacency matrix.
        """
        top_k_neigh = min(top_k_neigh, attention.size(-1))
        knn_val, knn_ind = torch.topk(attention, top_k_neigh, dim=-1)
        weighted_adj = to_cuda((self.mask_off_val * torch.ones_like(attention)).scatter_(-1, knn_ind, knn_val), self.device)

        return weighted_adj

    def _build_epsilon_neighbourhood(self, attention, epsilon_neigh):
        """Build epsilon neighbourhood graph.

        Parameters
        ----------
        attention : torch.Tensor
            The attention matrix.
        epsilon_neigh : float
            The threshold value for epsilon neighbourhood graph.

        Returns
        -------
        torch.Tensor
            The sparsified adjacency matrix.
        """
        mask = (attention > epsilon_neigh).detach().float()
        weighted_adj = attention * mask + self.mask_off_val * (1 - mask)

        return weighted_adj

    def _compute_distance_matrix(self, X, weight=None):
        """Compute distance matrix for RBF kernel.
        """
        if weight is not None:
            trans_X = torch.mm(X, weight)
        else:
            trans_X = X

        norm = torch.sum(trans_X * X, dim=-1)
        dists = -2 * torch.matmul(trans_X, X.transpose(-1, -2)) + norm.unsqueeze(0) + norm.unsqueeze(1)

        return dists

    def _get_node_mask_for_batch_graph(self, num_nodes):
        node_mask = []
        for i in range(num_nodes.shape[0]): # batch
            graph_node_num = num_nodes[i].item()
            node_mask.append(sparse.coo_matrix(np.ones((graph_node_num, graph_node_num))))

        node_mask = sparse.block_diag(node_mask)
        node_mask = to_cuda(sparse_mx_to_torch_sparse_tensor(node_mask).to_dense(), self.device)

        return node_mask

    def _get_normalized_init_adj(self, graph):
        adj = graph.scipy_sparse_adj()
        adj = normalize_sparse_adj(adj)
        adj = to_cuda(sparse_mx_to_torch_sparse_tensor(adj), self.device)

        return adj



from .graph_construction import get_graph_construction_args
from .graph_embedding import get_graph_embedding_args
from .prediction.generation import get_decoder_args


def get_basic_args(graph_construction_name, graph_embedding_name, decoder_name):
    """
        It will build the template for ``Graph2X`` model.
    Parameters
    ----------
    graph_construction_name: str
        The graph construction method name. Expected in ["dependency", "constituency", "node_emb", "node_emb_refined"].
    graph_embedding_name: str
        The graph embedding name. Expected in ["gcn", "gat", "graphsage", "ggnn"].
    decoder_name: str
        The decoder name. Expected in ["stdrnn", "stdtree"].

    Returns
    -------
    template_dict: dict
        The template dict.
        The structure is shown as follows:
            {
                graph_construction_args: dict,
                graph_embedding_args: dict,
                decoder_args: dict
            }
    """
    graph_construction_args = get_graph_construction_args(graph_construction_name)
    graph_embedding_args = get_graph_embedding_args(graph_embedding_name)
    decoder_args = get_decoder_args(decoder_name)
    ret = {
        "graph_construction_args": graph_construction_args,
        "graph_embedding_args": graph_embedding_args,
        "decoder_args": decoder_args
    }
    return ret


__all__ = ["get_graph_construction_args", "get_graph_embedding_args", "get_decoder_args", "get_basic_args"]



from graph4nlp.pytorch.modules.graph_construction.ie_graph_construction import IEBasedGraphConstruction
from stanfordcorenlp import StanfordCoreNLP
from graph4nlp.pytorch.modules.utils.vocab_utils import VocabModel


def test_ie_graph():
    # raw_data = "James went to the corner-shop. He want to buy some (eggs), <milk> and bread for breakfast."
    # raw_data = ('Return to Olympus is the only album by the alternative rock band Malfunkshun . '
    #             'It was released after the band had broken up and after lead singer Andrew Wood '
    #             '( later of Mother Love Bone ) had died of a drug overdose in 1990 . ')

    raw_data = ('James is on shop. He buys eggs.')

    # raw_data = ('Paul Allen was born on January 21, 1953, in Seattle, Washington, '
    #             'to Kenneth Sam Allen and Edna Faye Allen. Allen attended Lakeside School, '
    #             'a private school in Seattle, where he befriended Bill Gates, two years younger, '
    #             'with whom he shared an enthusiasm for computers. ')
                # 'Paul and Bill used a teletype terminal at their high school, Lakeside, '
                # 'to develop their programming skills on several time-sharing computer systems.')

    # raw_data = ('Police arrested a man named Albert Boucher . As of Saturday afternoon , he '
    #             'was being held in the Warren County Regional Jail on a $5,000 bond .')

    # use '_' replace '-' ? 'fourth-degree'->'fourth_degree'
    # java -mx4g -cp "*" edu.stanford.nlp.pipeline.StanfordCoreNLPServer -port 9000 -timeout 15000

    # raw_data = ('Rand Paul was assaulted in his home in Bowling Green, Kentucky , on Friday , '
    #             'according to Kentucky State Police . State troopers responded to a call to the senator \'s '
    #             'residence at 3:21 p.m. Friday . Police arrested a man named Rene Albert Boucher, who they '
    #             'allege " intentionally assaulted " Paul , causing him "minor injury" . Boucher , 59 , of Bowling '
    #             'Green was charged with one count of fourth_degree assault . As of Saturday afternoon , he '
    #             'was being held in the Warren County Regional Jail on a $5,000 bond .')

    nlp_parser = StanfordCoreNLP('http://localhost', port=9000, timeout=300000)

    # vocab_model =  VocabModel([raw_data.split('.'), raw_data.split('.')],
    #                     max_word_vocab_size=None,
    #                     min_word_vocab_freq=1,
    #                     word_emb_size=300)

    output_graph0 = IEBasedGraphConstruction.topology(raw_data, nlp_parser, merge_strategy=None, edge_strategy=None)

    # embedding_styles = {'word_emb_type':'w2v', 'node_edge_emb_strategy':'mean','seq_info_encode_strategy':'bilstm'}
    embedding_styles = {'word_emb_type':'w2v', 'node_edge_emb_strategy':'lstm','seq_info_encode_strategy':'bilstm'}
    # iegc = IEBasedGraphConstruction(vocab=vocab_model.word_vocab, embedding_style=embedding_styles, use_cuda=False)
    # output_graph0 = iegc.embedding(output_graph0)

    output_graph1 = IEBasedGraphConstruction.topology(raw_data, nlp_parser, merge_strategy="global", edge_strategy=None)
    # output_graph1 = iegc.embedding(output_graph1)

    output_graph2 = IEBasedGraphConstruction.topology(raw_data, nlp_parser, merge_strategy=None, edge_strategy='as_node')
    # output_graph2 = iegc.embedding(output_graph2)

    output_graph3 = IEBasedGraphConstruction.topology(raw_data, nlp_parser, merge_strategy="global", edge_strategy='as_node')
    # output_graph3 = iegc.embedding(output_graph3)

    # IEBasedGraphConstruction.topology(raw_data, nlp_parser, merge_strategy="share_common_tokens", edge_strategy='as_node')

    print(output_graph0.node_attributes)
    print(output_graph0.edge_attributes)

    print(output_graph1.node_attributes)
    print(output_graph1.edge_attributes)

    print(output_graph2.node_attributes)
    print(output_graph2.edge_attributes)

    print(output_graph3.node_attributes)
    print(output_graph3.edge_attributes)
    pass


if __name__ == "__main__":
    test_ie_graph()


import os
import torch

from ..data.dataset import DoubleText2TextDataset


class SQuADDataset(DoubleText2TextDataset):
    @property
    def raw_file_names(self):
        """3 reserved keys: 'train', 'val' (optional), 'test'. Represent the split of dataset."""
        return {'train': 'train.txt', 'val': 'val.txt', 'test': 'test.txt'}

    @property
    def processed_file_names(self):
        """At least 2 reserved keys should be fiiled: 'vocab' and 'data'."""
        return {'vocab': 'vocab.pt', 'data': 'data.pt'}

    def download(self):
        raise NotImplementedError(
            'This dataset is now under test and cannot be downloaded. Please prepare the raw data yourself.')

    def __init__(self, root_dir, topology_builder=None, topology_subdir=None, graph_type='static',
                 edge_strategy=None, merge_strategy='tailhead', max_word_vocab_size=None,
                 min_word_vocab_freq=1, word_emb_size=None, **kwargs):
        super(SQuADDataset, self).__init__(root_dir=root_dir, topology_builder=topology_builder,
                                          topology_subdir=topology_subdir, graph_type=graph_type,
                                          edge_strategy=edge_strategy, merge_strategy=merge_strategy,
                                          max_word_vocab_size=max_word_vocab_size,
                                          min_word_vocab_freq=min_word_vocab_freq,
                                          word_emb_size=word_emb_size, **kwargs)



from graph4nlp.pytorch.modules.evaluation.base import EvaluationMetricBase


class ExpressionAccuracy(EvaluationMetricBase):
    def __init__(self):
        super(ExpressionAccuracy, self).__init__()

    def calculate_scores(self, ground_truth, predict):
        correct = 0
        assert len(ground_truth) == len(predict)
        for gt, pred in zip(ground_truth, predict):
            if gt == pred:
                correct += 1.
        return correct / len(ground_truth)




import os

import torch
import pickle
import json
from graph4nlp.pytorch.data.dataset import Text2TextDataset, Text2TextDataItem
from graph4nlp.pytorch.modules.utils.vocab_utils import VocabModel
from ..modules.graph_construction.ie_graph_construction import IEBasedGraphConstruction
from ..modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
from ..modules.graph_construction.constituency_graph_construction import ConstituencyBasedGraphConstruction
import stanfordcorenlp
import time
dataset_root = '../../../examples/pytorch/summarization/cnn'


class CNNDataset(Text2TextDataset):
    @property
    def raw_file_names(self):
        """3 reserved keys: 'train', 'val' (optional), 'test'. Represent the split of dataset."""
        return {'train': 'train.json', 'test': 'test.json', 'val': 'val.json'}

    @property
    def processed_file_names(self):
        return {'vocab': 'vocab.pt', 'data': 'data.pt'}

    def download(self):
        # raise NotImplementedError(
        #     'This dataset is now under test and cannot be downloaded. Please prepare the raw data yourself.')
        return

    def __init__(self, root_dir, topology_builder=None, topology_subdir=None, graph_type='static',
                 edge_strategy=None, merge_strategy=None, **kwargs):
        super(CNNDataset, self).__init__(root_dir=root_dir, topology_builder=topology_builder,
                                          topology_subdir=topology_subdir, graph_type=graph_type,
                                          edge_strategy=edge_strategy, merge_strategy=merge_strategy, **kwargs)

    def build_topology(self, data_items):
        """
        Add exception process for IE-based graph construction. When the input text is too long,
        the processor may raise ``java.util.concurrent.TimeoutException``
        """
        if self.graph_type == 'static':
            print('Connecting to stanfordcorenlp server...')
            processor = stanfordcorenlp.StanfordCoreNLP('http://localhost', port=9098, timeout=1000)

            if self.topology_builder == IEBasedGraphConstruction:
                props_coref = {
                    'annotators': 'tokenize, ssplit, pos, lemma, ner, parse, coref',
                    "tokenize.options":
                        "splitHyphenated=true,normalizeParentheses=true,normalizeOtherBrackets=true",
                    "tokenize.whitespace": False,
                    'ssplit.isOneSentence': False,
                    'outputFormat': 'json'
                }
                props_openie = {
                    'annotators': 'tokenize, ssplit, pos, ner, parse, openie',
                    "tokenize.options":
                        "splitHyphenated=true,normalizeParentheses=true,normalizeOtherBrackets=true",
                    "tokenize.whitespace": False,
                    'ssplit.isOneSentence': False,
                    'outputFormat': 'json',
                    "openie.triple.strict": "true"
                }
                processor_args = [props_coref, props_openie]
            elif self.topology_builder == DependencyBasedGraphConstruction:
                processor_args = {
                    'annotators': 'ssplit,tokenize,depparse',
                    "tokenize.options":
                        "splitHyphenated=false,normalizeParentheses=false,normalizeOtherBrackets=false",
                    "tokenize.whitespace": False,
                    'ssplit.isOneSentence': False,
                    'outputFormat': 'json'
                }
            elif self.topology_builder == ConstituencyBasedGraphConstruction:
                processor_args = {
                    'annotators': "tokenize,ssplit,pos,parse",
                    "tokenize.options":
                    "splitHyphenated=true,normalizeParentheses=true,normalizeOtherBrackets=true",
                    "tokenize.whitespace": False,
                    'ssplit.isOneSentence': False,
                    'outputFormat': 'json'
                }
            else:
                raise NotImplementedError
            print('CoreNLP server connected.')
            pop_idxs = []
            for idx, item in enumerate(data_items):
                try:
                    graph = self.topology_builder.topology(raw_text_data=item.input_text,
                                                           nlp_processor=processor,
                                                           processor_args=processor_args,
                                                           merge_strategy=self.merge_strategy,
                                                           edge_strategy=self.edge_strategy,
                                                           verbase=False)
                    item.graph = graph
                except:
                    pop_idxs.append(idx)
                    item.graph = None
                    print('item does not have graph: '+str(idx))

            data_items = [x for idx, x in enumerate(data_items) if idx not in pop_idxs]
        elif self.graph_type == 'dynamic':
            if self.dynamic_graph_type == 'node_emb':
                for item in data_items:
                    graph = self.topology_builder.init_topology(item.input_text,
                                                                lower_case=self.lower_case,
                                                                tokenizer=self.tokenizer)
                    item.graph = graph
            elif self.dynamic_graph_type == 'node_emb_refined':
                if self.dynamic_init_topology_builder in (IEBasedGraphConstruction, DependencyBasedGraphConstruction, ConstituencyBasedGraphConstruction):
                    print('Connecting to stanfordcorenlp server...')
                    processor = stanfordcorenlp.StanfordCoreNLP('http://localhost', port=9000, timeout=1000)

                    if self.dynamic_init_topology_builder == IEBasedGraphConstruction:
                        props_coref = {
                            'annotators': 'tokenize, ssplit, pos, lemma, ner, parse, coref',
                            "tokenize.options":
                                "splitHyphenated=true,normalizeParentheses=true,normalizeOtherBrackets=true",
                            "tokenize.whitespace": False,
                            'ssplit.isOneSentence': False,
                            'outputFormat': 'json'
                        }
                        props_openie = {
                            'annotators': 'tokenize, ssplit, pos, ner, parse, openie',
                            "tokenize.options":
                                "splitHyphenated=true,normalizeParentheses=true,normalizeOtherBrackets=true",
                            "tokenize.whitespace": False,
                            'ssplit.isOneSentence': False,
                            'outputFormat': 'json',
                            "openie.triple.strict": "true"
                        }
                        processor_args = [props_coref, props_openie]
                    elif self.dynamic_init_topology_builder == DependencyBasedGraphConstruction:
                        processor_args = {
                            'annotators': 'ssplit,tokenize,depparse',
                            "tokenize.options":
                                "splitHyphenated=false,normalizeParentheses=false,normalizeOtherBrackets=false",
                            "tokenize.whitespace": False,
                            'ssplit.isOneSentence': False,
                            'outputFormat': 'json'
                        }
                    elif self.dynamic_init_topology_builder == ConstituencyBasedGraphConstruction:
                        processor_args = {
                            'annotators': "tokenize,ssplit,pos,parse",
                            "tokenize.options":
                            "splitHyphenated=true,normalizeParentheses=true,normalizeOtherBrackets=true",
                            "tokenize.whitespace": False,
                            'ssplit.isOneSentence': False,
                            'outputFormat': 'json'
                        }
                    else:
                        raise NotImplementedError
                    print('CoreNLP server connected.')
                else:
                    processor = None
                    processor_args = None

                for item in data_items:
                    graph = self.topology_builder.init_topology(item.input_text,
                                                                dynamic_init_topology_builder=self.dynamic_init_topology_builder,
                                                                lower_case=self.lower_case,
                                                                tokenizer=self.tokenizer,
                                                                nlp_processor=processor,
                                                                processor_args=processor_args,
                                                                merge_strategy=self.merge_strategy,
                                                                edge_strategy=self.edge_strategy,
                                                                verbase=False,
                                                                dynamic_init_topology_aux_args=self.dynamic_init_topology_aux_args)

                    item.graph = graph
            else:
                raise RuntimeError('Unknown dynamic_graph_type: {}'.format(self.dynamic_graph_type))

        else:
            raise NotImplementedError('Currently only static and dynamic are supported!')

    def parse_file(self, file_path) -> list:
        """
        Read and parse the file specified by `file_path`. The file format is specified by each individual task-specific
        base class. Returns all the indices of data items in this file w.r.t. the whole dataset.
        For Text2TextDataset, the format of the input file should contain lines of input, each line representing one
        record of data. The input and output is separated by a tab(\t).
        Examples
        --------
        input: list job use languageid0 job ( ANS ) , language ( ANS , languageid0 )
        DataItem: input_text="list job use languageid0", output_text="job ( ANS ) , language ( ANS , languageid0 )"
        Parameters
        ----------
        file_path: str
            The path of the input file.
        Returns
        -------
        list
            The indices of data items in the file w.r.t. the whole dataset.
        """
        data = []
        with open(file_path, 'r') as f:
            examples = json.load(f)
            for example_dict in examples:
                input = ' '.join(example_dict['article'][:10])
                output = example_dict['highlight'][0][0]
                data_item = Text2TextDataItem(input_text=input, output_text=output, tokenizer=self.tokenizer,
                                              share_vocab=self.share_vocab)
                data.append(data_item)
        return data


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--parser', type=str, default='IE')
    args = parser.parse_args()

    if args.parser=='IE':
        start_time = time.time()
        cnn_ie = CNNDataset(root_dir=dataset_root, topology_builder=IEBasedGraphConstruction,
                    topology_subdir='IEGraph')
        end_time = time.time() # 333.8594479560852
        # 141.86208820343018 [:10]
    elif args.parser=='DEP':
        start_time = time.time()
        cnn_dep = CNNDataset(root_dir=dataset_root, topology_builder=DependencyBasedGraphConstruction,
                        topology_subdir='DepGraph')
        end_time = time.time() # 22.60792326927185
    elif args.parser=='CONS':
        start_time = time.time()
        cnn_cons = CNNDataset(root_dir=dataset_root, topology_builder=ConstituencyBasedGraphConstruction,
                             topology_subdir='ConsGraph')
        end_time = time.time() # 48.007439851760864
    else:
        raise NotImplementedError()

    print(end_time - start_time)

    a = 0


from graph4nlp.pytorch.modules.evaluation.base import EvaluationMetricBase
from graph4nlp.pytorch.modules.evaluation.meteor_tool.meteor import Meteor


class METEOR(EvaluationMetricBase):
    """
        The METEOR evaluation metric class.
    Parameters
    ----------
    """

    def __init__(self):
        super(METEOR, self).__init__()
        self.scorer = Meteor()

    def calculate_scores(self, ground_truth, predict):
        """
            The METEOR calculation function. It will compute the METEOR scores.
        Parameters
        ----------
        ground_truth: list[string]
            The ground truth (correct) target values. It is a list of strings.
        predict: list[string]
            The predicted target values. It is a list of strings.
        Returns
        -------
        score: float
            The METEOR value.
        scores: list[float]
            The specific results for METEOR metric.
        """
        ref_list = [list(map(str.strip, refs)) for refs in zip(ground_truth)]

        refs = {idx: strippedlines for (idx, strippedlines) in enumerate(ref_list)}
        hyps = {idx: [lines.strip()] for (idx, lines) in enumerate(predict)}
        assert len(refs) == len(hyps)
        score, scores = self.scorer.compute_score(refs, hyps)
        return score, scores



import os
import time
import datetime
import argparse
import yaml

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import DataLoader
import torch.optim as optim
from torch.optim.lr_scheduler import ReduceLROnPlateau
import torch.backends.cudnn as cudnn

from graph4nlp.pytorch.datasets.trec import TrecDataset
from graph4nlp.pytorch.modules.graph_construction import *
from graph4nlp.pytorch.modules.graph_construction.embedding_construction import WordEmbedding
from graph4nlp.pytorch.modules.graph_embedding import GAT, GraphSAGE, GGNN
from graph4nlp.pytorch.modules.prediction.classification.graph_classification import FeedForwardNN
from graph4nlp.pytorch.modules.evaluation.base import EvaluationMetricBase
from graph4nlp.pytorch.modules.evaluation.accuracy import Accuracy
from graph4nlp.pytorch.modules.utils.generic_utils import grid, to_cuda, EarlyStopping
from graph4nlp.pytorch.modules.loss.general_loss import GeneralLoss
from graph4nlp.pytorch.modules.utils.logger import Logger
from graph4nlp.pytorch.modules.utils import constants as Constants
import torch.multiprocessing
torch.multiprocessing.set_sharing_strategy('file_system')



class TextClassifier(nn.Module):
    def __init__(self, vocab, config):
        super(TextClassifier, self).__init__()
        self.config = config
        self.vocab = vocab
        embedding_style = {'single_token_item': True if config['graph_type'] != 'ie' else False,
                            'emb_strategy': config.get('emb_strategy', 'w2v_bilstm'),
                            'num_rnn_layers': 1,
                            'bert_model_name': config.get('bert_model_name', 'bert-base-uncased'),
                            'bert_lower_case': True
                           }

        assert not (config['graph_type'] in ('node_emb', 'node_emb_refined') and config['gnn'] == 'gat'), \
                                'dynamic graph construction does not support GAT'

        use_edge_weight = False
        if config['graph_type'] == 'dependency':
            self.graph_topology = DependencyBasedGraphConstruction(embedding_style=embedding_style,
                                                                   vocab=vocab.in_word_vocab,
                                                                   hidden_size=config['num_hidden'],
                                                                   word_dropout=config['word_dropout'],
                                                                   rnn_dropout=config['rnn_dropout'],
                                                                   fix_word_emb=not config['no_fix_word_emb'],
                                                                   fix_bert_emb=not config.get('no_fix_bert_emb', False),
                                                                   device=config['device'])
        elif config['graph_type'] == 'constituency':
            self.graph_topology = ConstituencyBasedGraphConstruction(embedding_style=embedding_style,
                                                                   vocab=vocab.in_word_vocab,
                                                                   hidden_size=config['num_hidden'],
                                                                   word_dropout=config['word_dropout'],
                                                                   rnn_dropout=config['rnn_dropout'],
                                                                   fix_word_emb=not config['no_fix_word_emb'],
                                                                   fix_bert_emb=not config.get('no_fix_bert_emb', False),
                                                                   device=config['device'])
        elif config['graph_type'] == 'ie':
            self.graph_topology = IEBasedGraphConstruction(embedding_style=embedding_style,
                                                                   vocab=vocab.in_word_vocab,
                                                                   hidden_size=config['num_hidden'],
                                                                   word_dropout=config['word_dropout'],
                                                                   rnn_dropout=config['rnn_dropout'],
                                                                   fix_word_emb=not config['no_fix_word_emb'],
                                                                   fix_bert_emb=not config.get('no_fix_bert_emb', False),
                                                                   device=config['device'])
        elif config['graph_type'] == 'node_emb':
            self.graph_topology = NodeEmbeddingBasedGraphConstruction(
                                    vocab.in_word_vocab,
                                    embedding_style,
                                    sim_metric_type=config['gl_metric_type'],
                                    num_heads=config['gl_num_heads'],
                                    top_k_neigh=config['gl_top_k'],
                                    epsilon_neigh=config['gl_epsilon'],
                                    smoothness_ratio=config['gl_smoothness_ratio'],
                                    connectivity_ratio=config['gl_connectivity_ratio'],
                                    sparsity_ratio=config['gl_sparsity_ratio'],
                                    input_size=config['num_hidden'],
                                    hidden_size=config['gl_num_hidden'],
                                    fix_word_emb=not config['no_fix_word_emb'],
                                    fix_bert_emb=not config.get('no_fix_bert_emb', False),
                                    word_dropout=config['word_dropout'],
                                    rnn_dropout=config['rnn_dropout'],
                                    device=config['device'])
            use_edge_weight = True
        elif config['graph_type'] == 'node_emb_refined':
            self.graph_topology = NodeEmbeddingBasedRefinedGraphConstruction(
                                    vocab.in_word_vocab,
                                    embedding_style,
                                    config['init_adj_alpha'],
                                    sim_metric_type=config['gl_metric_type'],
                                    num_heads=config['gl_num_heads'],
                                    top_k_neigh=config['gl_top_k'],
                                    epsilon_neigh=config['gl_epsilon'],
                                    smoothness_ratio=config['gl_smoothness_ratio'],
                                    connectivity_ratio=config['gl_connectivity_ratio'],
                                    sparsity_ratio=config['gl_sparsity_ratio'],
                                    input_size=config['num_hidden'],
                                    hidden_size=config['gl_num_hidden'],
                                    fix_word_emb=not config['no_fix_word_emb'],
                                    fix_bert_emb=not config.get('no_fix_bert_emb', False),
                                    word_dropout=config['word_dropout'],
                                    rnn_dropout=config['rnn_dropout'],
                                    device=config['device'])
            use_edge_weight = True
        else:
            raise RuntimeError('Unknown graph_type: {}'.format(config['graph_type']))

        if 'w2v' in self.graph_topology.embedding_layer.word_emb_layers:
            self.word_emb = self.graph_topology.embedding_layer.word_emb_layers['w2v'].word_emb_layer
        else:
            self.word_emb = WordEmbedding(
                            self.vocab.in_word_vocab.embeddings.shape[0],
                            self.vocab.in_word_vocab.embeddings.shape[1],
                            pretrained_word_emb=self.vocab.in_word_vocab.embeddings,
                            fix_emb=not config['no_fix_word_emb'],
                            device=config['device']).word_emb_layer

        if config['gnn'] == 'gat':
            heads = [config['gat_num_heads']] * (config['gnn_num_layers'] - 1) + [config['gat_num_out_heads']]
            self.gnn = GAT(config['gnn_num_layers'],
                        config['num_hidden'],
                        config['num_hidden'],
                        config['num_hidden'],
                        heads,
                        direction_option=config['gnn_direction_option'],
                        feat_drop=config['gnn_dropout'],
                        attn_drop=config['gat_attn_dropout'],
                        negative_slope=config['gat_negative_slope'],
                        residual=config['gat_residual'],
                        activation=F.elu)
        elif config['gnn'] == 'graphsage':
            self.gnn = GraphSAGE(config['gnn_num_layers'],
                        config['num_hidden'],
                        config['num_hidden'],
                        config['num_hidden'],
                        config['graphsage_aggreagte_type'],
                        direction_option=config['gnn_direction_option'],
                        feat_drop=config['gnn_dropout'],
                        bias=True,
                        norm=None,
                        activation=F.relu,
                        use_edge_weight=use_edge_weight)
        elif config['gnn'] == 'ggnn':
            self.gnn = GGNN(config['gnn_num_layers'],
                        config['num_hidden'],
                        config['num_hidden'],
                        config['num_hidden'],
                        feat_drop=config['gnn_dropout'],
                        direction_option=config['gnn_direction_option'],
                        bias=True,
                        use_edge_weight=use_edge_weight)
        else:
            raise RuntimeError('Unknown gnn type: {}'.format(config['gnn']))

        self.clf = FeedForwardNN(2 * config['num_hidden'] if config['gnn_direction_option'] == 'bi_sep' else config['num_hidden'],
                        config['num_classes'],
                        [config['num_hidden']],
                        graph_pool_type=config['graph_pooling'],
                        dim=config['num_hidden'],
                        use_linear_proj=config['max_pool_linear_proj'])

        self.loss = GeneralLoss('CrossEntropy')


    def forward(self, graph_list, tgt=None, require_loss=True):
        # graph embedding construction
        batch_gd = self.graph_topology(graph_list)

        # run GNN
        self.gnn(batch_gd)

        # run graph classifier
        self.clf(batch_gd)
        logits = batch_gd.graph_attributes['logits']

        if require_loss:
            loss = self.loss(logits, tgt)
            return logits, loss
        else:
            return logits


class ModelHandler:
    def __init__(self, config):
        super(ModelHandler, self).__init__()
        self.config = config
        self.logger = Logger(self.config['out_dir'], config={k:v for k, v in self.config.items() if k != 'device'}, overwrite=True)
        self.logger.write(self.config['out_dir'])
        self._build_dataloader()
        self._build_model()
        self._build_optimizer()
        self._build_evaluation()

    def _build_dataloader(self):
        dynamic_init_topology_builder = None
        if self.config['graph_type'] == 'dependency':
            topology_builder = DependencyBasedGraphConstruction
            graph_type = 'static'
            merge_strategy = 'tailhead'
        elif self.config['graph_type'] == 'constituency':
            topology_builder = ConstituencyBasedGraphConstruction
            graph_type = 'static'
            merge_strategy = 'tailhead'
        elif self.config['graph_type'] == 'ie':
            topology_builder = IEBasedGraphConstruction
            graph_type = 'static'
            merge_strategy = 'global'
        elif self.config['graph_type'] == 'node_emb':
            topology_builder = NodeEmbeddingBasedGraphConstruction
            graph_type = 'dynamic'
            merge_strategy = None
        elif self.config['graph_type'] == 'node_emb_refined':
            topology_builder = NodeEmbeddingBasedRefinedGraphConstruction
            graph_type = 'dynamic'
            merge_strategy = 'tailhead'

            if self.config['init_graph_type'] == 'line':
                dynamic_init_topology_builder = None
            elif self.config['init_graph_type'] == 'dependency':
                dynamic_init_topology_builder = DependencyBasedGraphConstruction
            elif self.config['init_graph_type'] == 'constituency':
                dynamic_init_topology_builder = ConstituencyBasedGraphConstruction
            elif self.config['init_graph_type'] == 'ie':
                merge_strategy = 'global'
                dynamic_init_topology_builder = IEBasedGraphConstruction
            else:
                # dynamic_init_topology_builder
                raise RuntimeError('Define your own dynamic_init_topology_builder')
        else:
            raise RuntimeError('Unknown graph_type: {}'.format(self.config['graph_type']))

        topology_subdir = '{}_based_graph'.format(self.config['graph_type'])
        if self.config['graph_type'] == 'node_emb_refined':
            topology_subdir += '_{}'.format(self.config['init_graph_type'])

        dataset = TrecDataset(root_dir=self.config.get('root_dir', 'examples/pytorch/text_classification/data/trec'),
                              pretrained_word_emb_file=self.config['pre_word_emb_file'],
                              merge_strategy=merge_strategy,
                              seed=self.config['seed'],
                              device=config['device'],
                              thread_number=4,
                              port=9000,
                              timeout=15000,
                              word_emb_size=300,
                              graph_type=graph_type,
                              topology_builder=topology_builder,
                              topology_subdir=topology_subdir,
                              dynamic_graph_type=self.config['graph_type'] if self.config['graph_type'] in ('node_emb', 'node_emb_refined') else None,
                              dynamic_init_topology_builder=dynamic_init_topology_builder,
                              dynamic_init_topology_aux_args={'dummy_param': 0})

        self.train_dataloader = DataLoader(dataset.train, batch_size=self.config['batch_size'], shuffle=True,
                                           num_workers=self.config['num_workers'],
                                           collate_fn=dataset.collate_fn)
        if hasattr(dataset, 'val')==False:
            dataset.val = dataset.test
        self.val_dataloader = DataLoader(dataset.val, batch_size=self.config['batch_size'], shuffle=False,
                                          num_workers=self.config['num_workers'],
                                          collate_fn=dataset.collate_fn)
        self.test_dataloader = DataLoader(dataset.test, batch_size=self.config['batch_size'], shuffle=False,
                                          num_workers=self.config['num_workers'],
                                          collate_fn=dataset.collate_fn)
        self.vocab = dataset.vocab_model
        self.config['num_classes'] = dataset.num_classes
        self.num_train = len(dataset.train)
        self.num_val = len(dataset.val)
        self.num_test = len(dataset.test)
        print('Train size: {}, Val size: {}, Test size: {}'
            .format(self.num_train, self.num_val, self.num_test))
        self.logger.write('Train size: {}, Val size: {}, Test size: {}'
            .format(self.num_train, self.num_val, self.num_test))

    def _build_model(self):
        self.model = TextClassifier(self.vocab, self.config).to(self.config['device'])

    def _build_optimizer(self):
        parameters = [p for p in self.model.parameters() if p.requires_grad]
        self.optimizer = optim.Adam(parameters, lr=self.config['lr'])
        self.stopper = EarlyStopping(os.path.join(self.config['out_dir'], Constants._SAVED_WEIGHTS_FILE), patience=self.config['patience'])
        self.scheduler = ReduceLROnPlateau(self.optimizer, mode='max', factor=self.config['lr_reduce_factor'], \
            patience=self.config['lr_patience'], verbose=True)

    def _build_evaluation(self):
        self.metric = Accuracy(['accuracy'])

    def train(self):
        dur = []
        for epoch in range(self.config['epochs']):
            self.model.train()
            train_loss = []
            train_acc = []
            t0 = time.time()
            for i, data in enumerate(self.train_dataloader):
                graph_list, tgt = data
                graph_list = [to_cuda(g, self.config['device']) for g in graph_list]
                tgt = to_cuda(tgt, self.config['device'])
                logits, loss = self.model(graph_list, tgt, require_loss=True)
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
                train_loss.append(loss.item())

                pred = torch.max(logits, dim=-1)[1].cpu()
                train_acc.append(self.metric.calculate_scores(ground_truth=tgt.cpu(), predict=pred.cpu())[0])
                dur.append(time.time() - t0)

            val_acc = self.evaluate(self.val_dataloader)
            self.scheduler.step(val_acc)
            print('Epoch: [{} / {}] | Time: {:.2f}s | Loss: {:.4f} | Train Acc: {:.4f} | Val Acc: {:.4f}'.
              format(epoch + 1, self.config['epochs'], np.mean(dur), np.mean(train_loss), np.mean(train_acc), val_acc))
            self.logger.write('Epoch: [{} / {}] | Time: {:.2f}s | Loss: {:.4f} | Train Acc: {:.4f} | Val Acc: {:.4f}'.
                        format(epoch + 1, self.config['epochs'], np.mean(dur), np.mean(train_loss), np.mean(train_acc), val_acc))

            if self.stopper.step(val_acc, self.model):
                break

        return self.stopper.best_score

    def evaluate(self, dataloader):
        self.model.eval()
        with torch.no_grad():
            pred_collect = []
            gt_collect = []
            for i, data in enumerate(dataloader):
                graph_list, tgt = data
                graph_list = [to_cuda(g, self.config['device']) for g in graph_list]
                logits = self.model(graph_list, require_loss=False)
                pred_collect.append(logits)
                gt_collect.append(tgt)

            pred_collect = torch.max(torch.cat(pred_collect, 0), dim=-1)[1].cpu()
            gt_collect = torch.cat(gt_collect, 0).cpu()
            score = self.metric.calculate_scores(ground_truth=gt_collect, predict=pred_collect)[0]

            return score

    def test(self):
        # restored best saved model
        self.stopper.load_checkpoint(self.model)

        t0 = time.time()
        acc = self.evaluate(self.test_dataloader)
        dur = time.time() - t0
        print('Test examples: {} | Time: {:.2f}s |  Test Acc: {:.4f}'.
          format(self.num_test, dur, acc))
        self.logger.write('Test examples: {} | Time: {:.2f}s |  Test Acc: {:.4f}'.
          format(self.num_test, dur, acc))

        return acc


def main(config):
    # configure
    np.random.seed(config['seed'])
    torch.manual_seed(config['seed'])

    if not config['no_cuda'] and torch.cuda.is_available():
        print('[ Using CUDA ]')
        config['device'] = torch.device('cuda' if config['gpu'] < 0 else 'cuda:%d' % config['gpu'])
        torch.cuda.manual_seed(config['seed'])
        torch.cuda.manual_seed_all(config['seed'])
        torch.backends.cudnn.deterministic = True
        cudnn.benchmark = False
    else:
        config['device'] = torch.device('cpu')

    ts = datetime.datetime.now().timestamp()
    config['out_dir'] += '_{}'.format(ts)
    print('\n' + config['out_dir'])

    runner = ModelHandler(config)
    t0 = time.time()

    val_acc = runner.train()
    test_acc = runner.test()

    print('Removed best saved model file to save disk space')
    os.remove(runner.stopper.save_model_path)
    runtime = time.time() - t0
    print('Total runtime: {:.2f}s'.format(runtime))
    runner.logger.write('Total runtime: {:.2f}s\n'.format(runtime))
    runner.logger.close()

    return val_acc, test_acc


################################################################################
# ArgParse and Helper Functions #
################################################################################
def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-config', '--config', required=True, type=str, help='path to the config file')
    parser.add_argument('--grid_search', action='store_true', help='flag: grid search')
    args = vars(parser.parse_args())

    return args


def get_config(config_path='config.yml'):
    with open(config_path, 'r') as setting:
        config = yaml.load(setting)

    return config


def print_config(config):
    print('**************** MODEL CONFIGURATION ****************')
    for key in sorted(config.keys()):
        val = config[key]
        keystr = '{}'.format(key) + (' ' * (24 - len(key)))
        print('{} -->   {}'.format(keystr, val))
    print('**************** MODEL CONFIGURATION ****************')


def grid_search_main(config):
    grid_search_hyperparams = []
    log_path = config['out_dir']
    for k, v in config.items():
        if isinstance(v, list):
            grid_search_hyperparams.append(k)
            log_path += '_{}_{}'.format(k, v)

    logger = Logger(log_path, config=config, overwrite=True)

    best_config = None
    best_score = -1
    configs = grid(config)
    for cnf in configs:
        for k in grid_search_hyperparams:
            cnf['out_dir'] += '_{}_{}'.format(k, cnf[k])

        val_score, test_score = main(cnf)
        if best_score < test_score:
            best_score = test_score
            best_config = cnf
            print('Found a better configuration: {}'.format(best_score))
            logger.write('Found a better configuration: {}'.format(best_score))

    print('\nBest configuration:')
    logger.write('\nBest configuration:')
    for k in grid_search_hyperparams:
        print('{}: {}'.format(k, best_config[k]))
        logger.write('{}: {}'.format(k, best_config[k]))

    print('Best score: {}'.format(best_score))
    logger.write('Best score: {}\n'.format(best_score))
    logger.close()


if __name__ == '__main__':
    cfg = get_args()
    config = get_config(cfg['config'])
    print_config(config)
    if cfg['grid_search']:
        grid_search_main(config)
    else:
        main(config)

# screen -S ghn_code3  python -m  examples.pytorch.text_classification.run_text_classifier -config examples/pytorch/text_classification/config/CAirline/gat_undirected_constituency.yaml
# screen -S ghn_code1  python -m  examples.pytorch.text_classification.run_text_classifier -config examples/pytorch/text_classification/config/CAirline/graphsage_undirected_node_emb.yaml
# screen -S ghn_code1  python -m  examples.pytorch.text_classification.run_text_classifier -config examples/pytorch/text_classification/config/CAirline/gat_undirected_constituency.yaml
# screen -S ghn_code3 python -m  examples.pytorch.text_classification.run_text_classifier -config examples/pytorch/text_classification/config/CAirline/gat_dependency.yaml
# screen -S ghn_code1 python -m  examples.pytorch.text_classification.run_text_classifier -config examples/pytorch/text_classification/config/CAirline/gat_undirected_dependency.yaml --grid_search
# screen -S ghn_code1 python -m  examples.pytorch.text_classification.run_text_classifier -config examples/pytorch/text_classification/config/CAirline/gat_bi_sep_dependency.yaml --grid_search
# screen -S ghn_code1 python -m  examples.pytorch.text_classification.run_text_classifier -config examples/pytorch/text_classification/config/CNSST/gat_bi_sep_dependency.yaml --grid_search



#!/usr/bin/env python
#
# File Name : bleu_tool.py
#
# Description : Wrapper for BLEU scorer.
#
# Creation Date : 06-01-2015
# Last Modified : Thu 19 Mar 2015 09:13:28 PM PDT
# Authors : Hao Fang <hfang@uw.edu> and Tsung-Yi Lin <tl483@cornell.edu>

from .bleu_scorer import BleuScorer


class Bleu:
    def __init__(self, n=4, verbase=0):
        # default compute Blue score up to 4
        self._n = n
        self._hypo_for_image = {}
        self.ref_for_image = {}
        self.verbase = verbase

    def compute_score(self, gts, res):

        assert(gts.keys() == res.keys())
        imgIds = gts.keys()

        bleu_scorer = BleuScorer(n=self._n)
        for id in imgIds:
            hypo = res[id]
            ref = gts[id]

            # Sanity check.
            assert(type(hypo) is list)
            assert(len(hypo) == 1)
            assert(type(ref) is list)
            assert(len(ref) >= 1)

            bleu_scorer += (hypo[0], ref)

        #score, scores = bleu_scorer.compute_score(option='shortest')
        score, scores = bleu_scorer.compute_score(option='closest', verbose=self.verbase)
        #score, scores = bleu_scorer.compute_score(option='average', verbose=1)

        # return (bleu_tool, bleu_info)
        return score, scores

    def method(self):
        return "Bleu"



from graph4nlp.pytorch.modules.evaluation.base import EvaluationMetricBase
import pyrouge
import re
import random, os
import shutil


class SummarizationRouge(EvaluationMetricBase):
    def __init__(self):
        super(SummarizationRouge, self).__init__()
        self.rouge = pyrouge.Rouge155()

    def calculate_scores(self, ground_truth, predict):
        if os.path.exists('pred') and os.path.exists('gt'):
            shutil.rmtree("pred")
            shutil.rmtree("gt")
        os.makedirs("pred", exist_ok=False)
        os.makedirs("gt", exist_ok=False)
        cnt = 0
        for g, p in zip(ground_truth, predict):
            self.dump_tmp(g, os.path.join("gt", str(cnt) + ".txt"))
            self.dump_tmp(p, os.path.join("pred", str(cnt) + ".txt"))
            cnt += 1
        self.rouge.model_filename_pattern = '#ID#.txt'
        self.rouge.system_filename_pattern = '(\d+).txt'
        self.rouge.model_dir = "gt"
        self.rouge.system_dir = "pred"
        rouge_results = self.rouge.convert_and_evaluate()
        print(rouge_results)
        # dic = self.rouge.output_to_dict(rouge_results)
        # return dic


    def make_html_safe(self, s):
        """Replace any angled brackets in string s to avoid interfering with HTML attention visualizer."""
        s = s.replace("<", "&lt;")
        s = s.replace(">", "&gt;")
        return s

    def dump_tmp(self, content, path):
        output = "\n".join(content)
        output = self.make_html_safe(output)

        with open(path, "w") as f:
            f.write(output)

def read_file(path):
    with open(path, "r") as f:
        content = f.readlines()
    ret = []
    for line in content:
        if line.strip() == "":
            continue
        if line.split()[-1] != "</t>":
            line += " . </t>"
        sents = re.findall("<t>(.*?)</t>", line, flags=re.S|re.M)
        sen = [sent.strip() for sent in sents]
        ret.append(sen)
    return ret

if __name__ == "__main__":
    path = 'save_g2s_3w_copy_dep/'
    gt_file = "save_g2s_3w_copy_dep/cnn_tgt_output.txt"
    pred_file = "save_g2s_3w_copy_dep/cnn_pred_output.txt"
    gt_collect = read_file(gt_file)
    pred_collect = read_file(pred_file)
    assert len(gt_collect) == len(pred_collect)

    metric = SummarizationRouge().calculate_scores(ground_truth=gt_collect, predict=pred_collect)


from graph4nlp.pytorch.modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
from stanfordcorenlp import StanfordCoreNLP
from graph4nlp.pytorch.modules.utils.vocab_utils import VocabModel


def test_dependency():
    raw_data = "James went to the corner-shop. He want to buy some (eggs), <milk> and bread for breakfast."

    nlp_parser = StanfordCoreNLP('http://localhost', port=9000, timeout=300000)

    DependencyBasedGraphConstruction.topology(raw_data, nlp_parser, merge_strategy="tailhead",
                                              edge_strategy="heterogeneous", verbase=1)
    pass


if __name__ == "__main__":
    test_dependency()


import random
import unittest

import numpy as np
import torch
import tqdm
from sklearn import metrics

from graph4nlp.pytorch.modules.evaluation.accuracy import Accuracy

"""
    strict unittest by comparing the results with sklearn
"""


class MyTestCase(unittest.TestCase):
    def test_special_case(self):
        ground_truth = np.array([0, 1, 3, 1, 1])
        predict = np.array([0, 1, 3, 3, 1])
        mcm_sklearn = metrics.multilabel_confusion_matrix(ground_truth, predict)
        accuracy = Accuracy(metrics=["precision", "recall", "F1", "accuracy"])
        mcm_graph4ai = accuracy._calculate_confusion_matrix(ground_truth=ground_truth, predict=predict)

        self.assertEqual(True, (mcm_graph4ai == mcm_sklearn).all())
        print("mcm test passed")

        precision_sklearn, recall_sklearn, f1_sklearn, _ = metrics.precision_recall_fscore_support(y_true=ground_truth,
                                                                                                   y_pred=predict,
                                                                                                   average="micro")

        accuracy_sklearn = metrics.accuracy_score(y_true=ground_truth, y_pred=predict)

        precision_graph4ai, recall_graph4ai, f1_graph4ai, accuracy_graph4ai = accuracy.calculate_scores(
            ground_truth=torch.from_numpy(ground_truth), predict=torch.from_numpy(predict),
            average="micro")
        self.assertEqual(precision_graph4ai, precision_sklearn)
        self.assertEqual(recall_graph4ai, recall_sklearn)
        self.assertEqual(f1_graph4ai, f1_sklearn)
        self.assertEqual(accuracy_graph4ai, accuracy_sklearn)

        print("micro average special case test passed")

        accuracy_sklearn = metrics.accuracy_score(y_true=ground_truth, y_pred=predict)

        precision_sklearn, recall_sklearn, f1_sklearn, _ = metrics.precision_recall_fscore_support(y_true=ground_truth,
                                                                                                   y_pred=predict,
                                                                                                   average="macro")

        precision_graph4ai, recall_graph4ai, f1_graph4ai, accuracy_graph4ai = accuracy.calculate_scores(
            ground_truth=torch.from_numpy(ground_truth), predict=torch.from_numpy(predict),
            average="macro")
        self.assertEqual(precision_graph4ai, precision_sklearn)
        self.assertEqual(recall_graph4ai, recall_sklearn)
        self.assertEqual(f1_graph4ai, f1_sklearn)
        self.assertEqual(accuracy_graph4ai, accuracy_sklearn)
        print("macro average special case test passed")
        accuracy_sklearn = metrics.accuracy_score(y_true=ground_truth, y_pred=predict)
        precision_sklearn, recall_sklearn, f1_sklearn, _ = metrics.precision_recall_fscore_support(y_true=ground_truth,
                                                                                                   y_pred=predict,
                                                                                                   average="weighted")

        precision_graph4ai, recall_graph4ai, f1_graph4ai, accuracy_graph4ai = accuracy.calculate_scores(
            ground_truth=torch.from_numpy(ground_truth), predict=torch.from_numpy(predict),
            average="weighted")
        self.assertEqual(precision_graph4ai, precision_sklearn)
        self.assertEqual(recall_graph4ai, recall_sklearn)
        self.assertEqual(f1_graph4ai, f1_sklearn)
        self.assertEqual(accuracy_graph4ai, accuracy_sklearn)
        print("weighted average special case test passed")

    def test_random(self):
        accuracy = Accuracy(metrics=["accuracy", "precision", "recall", "F1"])
        for test_time in tqdm.tqdm(range(100)):
            length = random.randint(2, 1000000)
            n_labels = random.randint(2, 100000)
            ground_truth = []
            predict = []
            for i in range(length):
                ground_truth.append(random.randint(0, n_labels - 1))
                predict.append(random.randint(0, n_labels - 1))
            ground_truth = np.array(ground_truth)
            predict = np.array(predict)
            average = random.choice(["weighted", "macro", "micro"])
            accuracy_sklearn = metrics.accuracy_score(y_true=ground_truth, y_pred=predict)
            precision_sklearn, recall_sklearn, f1_sklearn, _ = metrics.precision_recall_fscore_support(
                y_true=ground_truth,
                y_pred=predict,
                average=average)

            accuracy_graph4ai, precision_graph4ai, recall_graph4ai, f1_graph4ai = accuracy.calculate_scores(
                ground_truth=torch.from_numpy(ground_truth), predict=torch.from_numpy(predict),
                average=average)
            self.assertEqual(precision_graph4ai, precision_sklearn)
            self.assertEqual(recall_graph4ai, recall_sklearn)
            self.assertEqual(f1_graph4ai, f1_sklearn)
            self.assertEqual(accuracy_graph4ai, accuracy_sklearn)
        print("random test 1 passed")

        for test_time in tqdm.tqdm(range(100)):
            length = random.randint(2, 1000000)
            n_labels = random.randint(2, 100000)
            replace_ratio = random.random() / 1.5
            ground_truth = []
            predict = []
            for i in range(length):
                ground_truth.append(random.randint(0, n_labels - 1))
                if random.random() < replace_ratio:
                    predict.append(random.randint(0, n_labels - 1))
                else:
                    predict.append(ground_truth[-1])
            ground_truth = np.array(ground_truth)
            predict = np.array(predict)
            average = random.choice(["weighted", "macro", "micro"])
            accuracy_sklearn = metrics.accuracy_score(y_true=ground_truth, y_pred=predict)
            precision_sklearn, recall_sklearn, f1_sklearn, _ = metrics.precision_recall_fscore_support(
                y_true=ground_truth,
                y_pred=predict,
                average=average)

            accuracy_graph4ai, precision_graph4ai, recall_graph4ai, f1_graph4ai = accuracy.calculate_scores(
                ground_truth=torch.from_numpy(ground_truth), predict=torch.from_numpy(predict),
                average=average)
            self.assertEqual(precision_graph4ai, precision_sklearn)
            self.assertEqual(recall_graph4ai, recall_sklearn)
            self.assertEqual(f1_graph4ai, f1_sklearn)
            self.assertEqual(accuracy_graph4ai, accuracy_sklearn)
        print("random test 2 passed")


if __name__ == '__main__':
    unittest.main()



import torch
import torch.nn as nn

from graph4nlp.pytorch.modules.utils.vocab_utils import Vocab


class Graph2seqLoss(nn.Module):
    def __init__(self, vocab: Vocab):
        super(Graph2seqLoss, self).__init__()
        self.loss_func = nn.NLLLoss()
        self.VERY_SMALL_NUMBER = 1e-31
        self.vocab = vocab

    def forward(self, prob, gt):
        assert prob.shape[0:1] == gt.shape[0:1]
        assert len(prob.shape) == 3
        log_prob = torch.log(prob + self.VERY_SMALL_NUMBER)
        batch_size = gt.shape[0]
        step = gt.shape[1]

        mask = 1 - gt.data.eq(self.vocab.PAD).float()

        prob_select = torch.gather(log_prob.view(batch_size * step, -1), 1, gt.view(-1, 1))

        prob_select_masked = - torch.masked_select(prob_select, mask.view(-1, 1).bool())
        loss = torch.mean(prob_select_masked)
        return loss


class CoverageLoss(nn.Module):
    def __init__(self, cover_loss):
        super(CoverageLoss, self).__init__()
        self.cover_loss = cover_loss

    def forward(self, batch_size, enc_attn_weights, coverage_vectors):
        target_length = len(enc_attn_weights)
        loss = 0
        for i in range(target_length):
            if coverage_vectors[i] is not None:
                coverage_loss = torch.sum(
                    torch.min(coverage_vectors[i], enc_attn_weights[i])) / batch_size * self.cover_loss
                loss += coverage_loss
        return loss / target_length



from ...data.data import GraphData, from_batch, to_batch
import random
import torch

num_nodes = 50
num_edges = 50
num_graphs = 10

graph_list = []

for i in range(num_graphs):
    g = GraphData()
    g.add_nodes(num_nodes)
    src_list, tgt_list = [], []

    for i in range(num_edges):
        src = random.randint(a=0, b=num_nodes - 1)
        tgt = random.randint(a=0, b=num_nodes - 1)
        while tgt == src:
            tgt = random.randint(a=0, b=num_nodes - 1)
        src_list.append(src)
        tgt_list.append(tgt)

    for i in range(num_nodes):
        g.node_attributes[i]['name'] = '*NAMEFIX_{}'.format(i)
        g.node_attributes[i]['value'] = i
    g.add_edges(src_list, tgt_list)

    g.node_features['node_feat'] = torch.rand((50, 10))
    g.edge_features['edge_feat'] = torch.rand((50, 10))

    graph_list.append(g)

batch = to_batch(graph_list)
graph_list_2 = from_batch(batch)
batch.adj_matrix()
batch.scipy_sparse_adj()
eyes = torch.eye(10)
g1 = GraphData()
g1.from_dense_adj(eyes)
import numpy as np
a = np.array(list(range(10)))
import scipy.sparse.coo
coo = scipy.sparse.coo.coo_matrix((np.ones(10), (a, a)), shape=(10, 10))
g2 = GraphData()
g2.from_scipy_sparse_matrix(coo)
pass



from .dependency_graph_construction import DependencyBasedGraphConstruction
from .constituency_graph_construction import ConstituencyBasedGraphConstruction
from .ie_graph_construction import IEBasedGraphConstruction
from .node_embedding_based_graph_construction import NodeEmbeddingBasedGraphConstruction
from .node_embedding_based_refined_graph_construction import NodeEmbeddingBasedRefinedGraphConstruction

__all__ = ['DependencyBasedGraphConstruction',
            'ConstituencyBasedGraphConstruction',
            'IEBasedGraphConstruction',
            'NodeEmbeddingBasedGraphConstruction',
            'NodeEmbeddingBasedRefinedGraphConstruction']



from scipy import sparse

from ...data.data import GraphData

CORENLP_TIMEOUT_SIGNATURE = 'CoreNLP request timed out. Your document may be too long.'


def convert_adj_to_graph(adj, reverse_adj, mask_off_val):
    """Convert adjacency matrix to GraphData
    """
    binarized_adj = sparse.coo_matrix(adj.detach().cpu().numpy() != mask_off_val)
    graph_data = GraphData()
    graph_data.from_scipy_sparse_matrix(binarized_adj)
    edge_weight = adj[adj != mask_off_val]
    reverse_edge_weight = reverse_adj[reverse_adj != mask_off_val]

    graph_data.edge_features['edge_weight'] = edge_weight
    graph_data.edge_features['reverse_edge_weight'] = reverse_edge_weight

    return graph_data



from graph4nlp.pytorch.data.dataset import Text2TextDataItem, Text2TextDataset
from graph4nlp.pytorch.modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.constituency_graph_construction import ConstituencyBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.ie_graph_construction import IEBasedGraphConstruction
import torch
import os
import json
import stanfordcorenlp
import warnings
from multiprocessing import Pool
import numpy as np
from graph4nlp.pytorch.modules.utils.padding_utils import pad_2d_vals_no_size

from multiprocessing import Process
import multiprocessing
import tqdm
from graph4nlp.pytorch.modules.utils.vocab_utils import VocabModel, Vocab
from graph4nlp.pytorch.modules.utils import constants
from nltk.tokenize import word_tokenize

class CNNDataset(Text2TextDataset):
    def __init__(self,
                 root_dir,
                 topology_builder,
                 topology_subdir,
                 tokenizer=word_tokenize,
                 lower_case=True,
                 pretrained_word_emb_file=None,
                 use_val_for_vocab=False,
                 seed=1234,
                 device='cpu',
                 thread_number=4,
                 port=9000,
                 timeout=15000,
                 graph_type='static',
                 edge_strategy=None,
                 merge_strategy='tailhead',
                 share_vocab=True,
                 word_emb_size=300,
                 dynamic_graph_type=None,
                 dynamic_init_topology_builder=None,
                 dynamic_init_topology_aux_args=None
                 ):
        super(CNNDataset, self).__init__(root_dir=root_dir,
                                         topology_builder=topology_builder,
                                         topology_subdir=topology_subdir,
                                         tokenizer=tokenizer,
                                         lower_case=lower_case,
                                         pretrained_word_emb_file=pretrained_word_emb_file,
                                         use_val_for_vocab=use_val_for_vocab,
                                         seed=seed,
                                         device=device,
                                         thread_number=thread_number,
                                         port=port,
                                         timeout=timeout,
                                         graph_type=graph_type,
                                         edge_strategy=edge_strategy,
                                         merge_strategy=merge_strategy,
                                         share_vocab=share_vocab,
                                         word_emb_size=word_emb_size,
                                         dynamic_graph_type=dynamic_graph_type,
                                         dynamic_init_topology_builder=dynamic_init_topology_builder,
                                         dynamic_init_topology_aux_args=dynamic_init_topology_aux_args)

    @property
    def raw_file_names(self):
        """3 reserved keys: 'train', 'val' (optional), 'test'. Represent the split of dataset."""
        # return {'train': 'train_300.json', 'val': "train_30.json", 'test': 'train_30.json'}
        # return {'train': 'train_1w.json', 'val': "val.json", 'test': 'test.json'}
        return {'train': 'train_3w.json', 'val': "val.json", 'test': 'test.json'}

    @property
    def processed_file_names(self):
        """At least 3 reserved keys should be fiiled: 'vocab', 'data' and 'split_ids'."""
        return {'vocab': 'vocab.pt', 'data': 'data.pt'}

    def download(self):
        return

    def build_vocab(self):
        data_for_vocab = self.train
        if self.use_val_for_vocab:
            data_for_vocab = data_for_vocab + self.val

        vocab_model = VocabModel.build(saved_vocab_file=self.processed_file_paths['vocab'],
                                       data_set=data_for_vocab,
                                       tokenizer=self.tokenizer,
                                       lower_case=self.lower_case,
                                       max_word_vocab_size=None,
                                       min_word_vocab_freq=3,
                                       pretrained_word_emb_file=self.pretrained_word_emb_file,
                                       word_emb_size=self.word_emb_size,
                                       share_vocab=self.share_vocab)
        self.vocab_model = vocab_model
        return self.vocab_model

    def parse_file(self, file_path):
        """
        Read and parse the file specified by `file_path`. The file format is specified by each individual task-specific
        base class. Returns all the indices of data items in this file w.r.t. the whole dataset.
        For Text2TextDataset, the format of the input file should contain lines of input, each line representing one
        record of data. The input and output is separated by a tab(\t).
        Examples
        --------
        input: list job use languageid0 job ( ANS ) , language ( ANS , languageid0 )
        DataItem: input_text="list job use languageid0", output_text="job ( ANS ) , language ( ANS , languageid0 )"
        Parameters
        ----------
        file_path: str
            The path of the input file.
        Returns
        -------
        list
            The indices of data items in the file w.r.t. the whole dataset.
        """
        data = []
        with open(file_path, 'r') as f:
            examples = json.load(f)
            for example_dict in examples:
                input = ' '.join(' '.join(example_dict['article']).split()[:400]).lower()
                output = ' '.join(' '.join(['<t> ' + sent[0] + ' . </t>' for sent in example_dict['highlight']]).split()[:99]).lower()
                if input=='' or output=='':
                    continue
                data_item = Text2TextDataItem(input_text=input, output_text=output, tokenizer=self.tokenizer,
                                              share_vocab=self.share_vocab)
                data.append(data_item)
        return data



import torch

from ...modules.graph_construction.embedding_construction import EmbeddingConstruction
from ...modules.utils.vocab_utils import VocabModel
from ...modules.utils.padding_utils import pad_2d_vals_no_size


if __name__ == '__main__':
    raw_text_data = [['I like nlp.', 'Same here!'], ['I like graph.', 'Same here!']]

    vocab_model = VocabModel(raw_text_data, max_word_vocab_size=None,
                                min_word_vocab_freq=1,
                                word_emb_size=300)

    src_text_seq = list(zip(*raw_text_data))[0]
    src_idx_seq = [vocab_model.word_vocab.to_index_sequence(each) for each in src_text_seq]
    src_len = torch.LongTensor([len(each) for each in src_idx_seq])
    num_seq = torch.LongTensor([len(src_len)])
    input_tensor = torch.LongTensor(pad_2d_vals_no_size(src_idx_seq))
    print('input_tensor: {}'.format(input_tensor.shape))

    emb_constructor = EmbeddingConstruction(vocab_model.word_vocab, 'w2v', 'bilstm', 'bilstm', 128)
    emb = emb_constructor(input_tensor, src_len, num_seq)
    print('emb: {}'.format(emb.shape))



from torch import nn
import torch
from ..base import KGCompletionLayerBase


class ComplExLayer(KGCompletionLayerBase):
    r"""Specific class for knowledge graph completion task.

        ComplEx from paper `Complex Embeddings for Simple Link Prediction
        <http://proceedings.mlr.press/v48/trouillon16.pdf>`__.

        Parameters
        ----------
        input_dropout: float
           Dropout for node_emb and rel_emb. Default: 0.0

        rel_emb_from_gnn: bool
           If `rel_emb` is computed from GNN, rel_emb_from_gnn is set to `True`.
           Else, rel_emb is initialized as nn.Embedding randomly. Default: `True`.

        num_relations: int
           Number of relations. `num_relations` is needed if rel_emb_from_gnn==True.
           Default: `None`.

        embedding_dim: int
           Dimension of the rel_emb. `embedding_dim` is needed if rel_emb_from_gnn==True.
           Default: `0`.

        loss_name: str
           The loss type selected fot the KG completion task.

    """
    def __init__(self,
                 input_dropout=0.0,
                 rel_emb_from_gnn=True,
                 num_relations=None,
                 embedding_dim=None,
                 loss_name='BCELoss'):
        super(ComplExLayer, self).__init__()
        self.rel_emb_from_gnn = rel_emb_from_gnn
        self.inp_drop = nn.Dropout(input_dropout)
        if self.rel_emb_from_gnn == False:
            assert num_relations != None
            assert embedding_dim != None
            self.emb_rel_real = torch.nn.Embedding(num_relations, embedding_dim)
            self.emb_rel_img = torch.nn.Embedding(num_relations, embedding_dim)
            self.reset_parameters()
        self.loss_name = loss_name
        self.reset_parameters()


    def reset_parameters(self):
        if self.rel_emb_from_gnn == False:
            nn.init.xavier_normal_(self.emb_rel_real.weight.data)
            nn.init.xavier_normal_(self.emb_rel_img.weight.data)


    def forward(self,
                node_emb_real,
                node_emb_img,
                rel_emb_real=None,
                rel_emb_img=None,
                list_e_r_pair_idx=None,
                list_e_e_pair_idx=None,
                multi_label=None):
        r"""

        Parameters
        ----------

        node_emb_real: tensor [N, H]
            N: number of nodes in the whole KG graph
            H: length of the node embeddings (entity embeddings)

        node_emb_img: tensor [N, H]
            N: number of nodes in the whole KG graph
            H: length of the node embeddings (entity embeddings)

        rel_emb_real: tensor [N_r,H]
            N_r: number of relations in the whole KG graph
            H: length of the relation embeddings

        rel_emb_img: tensor [N_r,H]
            N_r: number of relations in the whole KG graph
            H: length of the relation embeddings

        list_e_r_pair_idx: list of tuple
            a list of index of head entities and relations that needs
            predicting the tail entities between them. Default: `None`

        list_e_e_pair_idx: list of tuple
            a list of index of head entities and tail entities that needs
            predicting the relations between them. Default: `None`.
            Only one of `list_e_r_pair_idx` and `list_e_e_pair_idx` can be `None`.

        multi_label: tensor [L, N]
            multi_label is a binary matrix. Each element can be equal to 1 for true label
            and 0 for false label (or 1 for true label, -1 for false label).
            multi_label[i] represents a multi-label of a given head-rel pair or head-tail pair.
            L is the length of list_e_r_pair_idx, list_e_e_pair_idx or batch size.
            N: number of nodes in the whole KG graph.

        Returns
        -------
        pred: tensor [L, N]
            logit tensor. [L, N] The score logits for all nodes preidcted.

        pred_pos: tensor [L_p]
            The predition scores of positive examples.

        pred_neg: tensor [L_n]
            The predition scores of negative examples. L_p + L_n == L * N.

        """
        if self.rel_emb_from_gnn == False:
            assert rel_emb_real is None
            assert rel_emb_img is None
            rel_emb_real = self.emb_rel_real.weight
            rel_emb_img = self.emb_rel_img.weight

        if list_e_r_pair_idx is None and list_e_e_pair_idx is None:
            raise RuntimeError("Only one of `list_e_r_pair_idx` and `list_e_e_pair_idx` can be `None`.")

        assert node_emb_real.size()[1]==rel_emb_real.size()[1]
        assert rel_emb_img.size()[1]==rel_emb_img.size()[1]

        if list_e_r_pair_idx is not None:
            ent_idxs = torch.LongTensor([x[0] for x in list_e_r_pair_idx])
            rel_idxs = torch.LongTensor([x[1] for x in list_e_r_pair_idx])

            selected_ent_embs_real = node_emb_real[ent_idxs].squeeze()  # [L, H]. L is the length of list_e_r_pair_idx
            selected_ent_embs_img = node_emb_img[ent_idxs].squeeze()  # [L, H]. L is the length of list_e_r_pair_idx
            selected_rel_embs_real = rel_emb_real[rel_idxs].squeeze()  # [L, H]. L is the length of list_e_r_pair_idx
            selected_rel_embs_img = rel_emb_img[rel_idxs].squeeze()  # [L, H]. L is the length of list_e_r_pair_idx

            # dropout
            selected_ent_embs_real = self.inp_drop(selected_ent_embs_real)
            selected_ent_embs_img = self.inp_drop(selected_ent_embs_img)
            selected_rel_embs_real = self.inp_drop(selected_rel_embs_real)
            selected_rel_embs_img = self.inp_drop(selected_rel_embs_img)

            # complex space bilinear product (equivalent to HolE)
            realrealreal = torch.mm(selected_ent_embs_real * selected_rel_embs_real,
                                    node_emb_real.transpose(1, 0))
            realimgimg = torch.mm(selected_ent_embs_real * selected_rel_embs_img,
                                  node_emb_img.transpose(1, 0))
            imgrealimg = torch.mm(selected_ent_embs_img * selected_rel_embs_real,
                                  node_emb_img.transpose(1, 0))
            imgimgreal = torch.mm(selected_ent_embs_img * selected_rel_embs_img,
                                  node_emb_real.transpose(1, 0))
            pred = realrealreal + realimgimg + imgrealimg - imgimgreal
        elif list_e_e_pair_idx is not None:
            raise NotImplementedError()

        if self.loss_name in ["SoftMarginLoss"]:
            pred = torch.tanh(pred)
        else:
            pred = torch.sigmoid(pred)

        if multi_label is not None:
            idxs_pos = torch.nonzero(multi_label == 1.)
            pred_pos = pred[idxs_pos[:, 0], idxs_pos[:, 1]]

            idxs_neg = torch.nonzero(multi_label == 0.)
            pred_neg = pred[idxs_neg[:, 0], idxs_neg[:, 1]]

            return pred, pred_pos, pred_neg
        else:
            return pred


from graph4nlp.pytorch.data.data import from_batch, GraphData
from graph4nlp.pytorch.modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.constituency_graph_construction import ConstituencyBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.node_embedding_based_graph_construction import NodeEmbeddingBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.node_embedding_based_refined_graph_construction import NodeEmbeddingBasedRefinedGraphConstruction

from graph4nlp.pytorch.modules.graph_embedding.gat import GAT
from graph4nlp.pytorch.modules.graph_embedding.ggnn import GGNN
from graph4nlp.pytorch.modules.graph_embedding.graphsage import GraphSAGE
from graph4nlp.pytorch.modules.graph_embedding.gcn import GCN
from graph4nlp.pytorch.modules.prediction.generation.StdRNNDecoder import StdRNNDecoder
from .loss import *
import torch.nn.functional as F


class Graph2seq(nn.Module):
    def __init__(self, vocab, hidden_size=300, graph_type="dependency", direction_option='undirected',
                 gnn="GAT", decoder_length=100, use_copy=False, use_coverage=False, graph_pooling_strategy=None,
                 attention_type="sep_diff_encoder_type", fuse_strategy="concatenate",
                 device=None, emb_dropout=0.2, feats_dropout=0.2, rnn_dropout=0.3):
        super(Graph2seq, self).__init__()

        self.vocab = vocab
        self.use_copy = use_copy
        self.use_coverage = use_coverage
        self.graph_type = graph_type
        use_edge_weight = False

        embedding_style = {'single_token_item': True,
                           'emb_strategy': "w2v_bilstm",
                           'num_rnn_layers': 1
                           }
        if self.graph_type == "dependency":
            self.graph_topology = DependencyBasedGraphConstruction(embedding_style=embedding_style,
                                                                   vocab=vocab.in_word_vocab,
                                                                   hidden_size=hidden_size,
                                                                   word_dropout=emb_dropout,
                                                                   rnn_dropout=rnn_dropout, device=device,
                                                                   fix_word_emb=False)
        elif self.graph_type == "constituency":
            self.graph_topology = ConstituencyBasedGraphConstruction(embedding_style=embedding_style,
                                                                     vocab=vocab.in_word_vocab,
                                                                     hidden_size=hidden_size, device=device,
                                                                     word_dropout=emb_dropout,
                                                                     rnn_dropout=rnn_dropout,
                                                                     fix_word_emb=False)
        elif self.graph_type == "node_emb":
            self.graph_topology = NodeEmbeddingBasedGraphConstruction(
                vocab.in_word_vocab,
                embedding_style,
                sim_metric_type='weighted_cosine',
                num_heads=1,
                top_k_neigh=None,
                epsilon_neigh=0.5,
                smoothness_ratio=0.1,
                connectivity_ratio=0.05,
                sparsity_ratio=0.1,
                input_size=hidden_size,
                hidden_size=hidden_size,
                fix_word_emb=False,
                word_dropout=emb_dropout,
                rnn_dropout=rnn_dropout,
                device=device)
            use_edge_weight = True
        elif self.graph_type == "node_emb_refined":
            self.graph_topology = NodeEmbeddingBasedRefinedGraphConstruction(
                vocab.in_word_vocab,
                embedding_style,
                0.2,
                sim_metric_type="weighted_cosine",
                num_heads=1,
                top_k_neigh=None,
                epsilon_neigh=0.5,
                smoothness_ratio=0.1,
                connectivity_ratio=0.05,
                sparsity_ratio=0.1,
                input_size=hidden_size,
                hidden_size=hidden_size,
                fix_word_emb=False,
                word_dropout=emb_dropout,
                rnn_dropout=rnn_dropout,
                device=device)
            use_edge_weight = True
        else:
            raise NotImplementedError()
        self.word_emb = self.graph_topology.embedding_layer.word_emb_layers['w2v'].word_emb_layer

        if gnn == "GAT":
            self.gnn_encoder = GAT(3, hidden_size, hidden_size, hidden_size, [2, 2, 1],
                                   direction_option=direction_option,
                                   feat_drop=feats_dropout, attn_drop=feats_dropout, activation=F.relu, residual=True)
        elif gnn == "GGNN":
            self.gnn_encoder = GGNN(3, hidden_size, hidden_size, direction_option=direction_option,
                                    use_edge_weight=use_edge_weight, feat_drop=feats_dropout)
        elif gnn == "GraphSage":
            self.gnn_encoder = GraphSAGE(3, hidden_size, hidden_size, hidden_size, aggregator_type="lstm",
                                         direction_option=direction_option, feat_drop=feats_dropout,
                                         activation=nn.ReLU(), bias=True, use_edge_weight=use_edge_weight)
        elif gnn == "GCN":
            self.gnn_encoder = GCN(3, hidden_size, hidden_size, hidden_size,direction_option=direction_option,
                                allow_zero_in_degree=True, activation=F.relu)
        else:
            raise NotImplementedError()

        self.seq_decoder = StdRNNDecoder(max_decoder_step=decoder_length,
                                         input_size=2*hidden_size if direction_option == 'bi_sep' else hidden_size,
                                         hidden_size=hidden_size, graph_pooling_strategy=graph_pooling_strategy,
                                         word_emb=self.word_emb, vocab=self.vocab.in_word_vocab,
                                         attention_type=attention_type, fuse_strategy=fuse_strategy,
                                         rnn_emb_input_size=hidden_size, use_coverage=use_coverage, use_copy=use_copy,
                                         tgt_emb_as_output_layer=True, dropout=rnn_dropout)
        self.loss_calc = Graph2seqLoss(self.vocab.in_word_vocab)
        self.loss_cover = CoverageLoss(0.3)

    @classmethod
    def from_args(cls, vocab, args, device):
        return cls(vocab=vocab, hidden_size=args.hidden_size, graph_type=args.graph_type,
                   direction_option=args.gnn_direction, gnn=args.gnn, device=device,
                   emb_dropout=args.emb_dropout, feats_dropout=args.feats_dropout, rnn_dropout=args.feats_dropout,
                   use_copy=args.use_copy, use_coverage=args.use_coverage, decoder_length=args.decoder_length,
                   graph_pooling_strategy=args.graph_pooling_strategy, attention_type=args.attention_type,
                   fuse_strategy=args.fuse_strategy)

    def forward(self, graph_list, tgt=None, oov_dict=None, require_loss=True):
        batch_graph = self.graph_topology(graph_list)
        # run GNN
        batch_graph: GraphData = self.gnn_encoder(batch_graph)
        batch_graph.node_features["rnn_emb"] = batch_graph.node_features['node_feat']
        graph_list_decoder = from_batch(batch_graph)
        if self.use_copy and "token_id_oov" not in batch_graph.node_features.keys():
            for g, g_ori in zip(graph_list_decoder, graph_list):
                g.node_features['token_id_oov'] = g_ori.node_features['token_id_oov']

        # down-task
        prob, enc_attn_weights, coverage_vectors = self.seq_decoder(graph_list_decoder, tgt_seq=tgt, oov_dict=oov_dict)
        if require_loss:
            loss = self.loss_calc(prob, tgt)
            if self.use_coverage:
                cover_loss = self.loss_cover(prob.shape[0], enc_attn_weights, coverage_vectors)
                loss += cover_loss
            return prob, loss
        else:
            return prob



from ..base import KGCompletionBase
from .TransELayer import TransELayer
from .....data.data import GraphData


class TransE(KGCompletionBase):
    r"""Specific class for knowledge graph completion task.
    TransE from paper `Translating Embeddings for Modeling
    Multi-relational Data <https://papers.nips.cc/paper/5071
    -translating-embeddings-for-modeling-multi-relational-data.pdf>`__.

    .. math::
        f(s, r, o) & = ||e_s + w_r - e_o||_p

    Parameters
    ----------
    p_norm: int
        Default: 1

    rel_emb_from_gnn: bool
        If `rel_emb` is computed from GNN, rel_emb_from_gnn is set to `True`.
        Else, rel_emb is initialized as nn.Embedding randomly. Default: `True`.

    num_relations: int
        Number of relations. `num_relations` is needed if rel_emb_from_gnn==True.
        Default: `None`.

    embedding_dim: int
        Dimension of the rel_emb. `embedding_dim` is needed if rel_emb_from_gnn==True.
        Default: `0`.

    edge2node: bool
        The edges or relations in KG are converted to nodes. Defauly: `False`.

    loss_name: str
        The loss type selected fot the KG completion task. Default: `'BCELoss'`
    """

    def __init__(self,
                 p_norm=1,
                 rel_emb_from_gnn=True,
                 num_relations=None,
                 embedding_dim=None,
                 edge2node=False,
                 loss_name='BCELoss'):
        super(TransE, self).__init__()
        self.rel_emb_from_gnn = rel_emb_from_gnn
        self.edge2node = edge2node
        self.classifier = TransELayer(p_norm, rel_emb_from_gnn,
                                       num_relations, embedding_dim, loss_name)


    def forward(self, input_graph: GraphData):
        r"""
        Forward functions to compute the logits tensor for kg completion.

        Parameters
        ----------

        input graph : GraphData
                     The tensors stored in the node feature field named "node_emb" and
                     "rel_emb" in the input_graph are used for knowledge graph completion.

        Returns
        ---------

        output_graph : GraphData
                      The computed logit tensor for each nodes in the graph are stored
                      in the node feature field named "node_logits".
                      logit tensor shape is: [num_class]
        """

        node_emb = input_graph.node_features['node_emb']
        if self.loss_name in ['SoftplusLoss', 'SigmoidLoss']:
            multi_label = input_graph.graph_attributes['multi_binary_label']
        else:
            multi_label = None

        if self.edge2node:
            rel_emb = node_emb
        else:
            if input_graph.edge_features['edge_emb'] != None:
                rel_emb = input_graph.edge_features['edge_emb']
            else:
                assert self.rel_emb_from_gnn == False
                rel_emb = None

        if 'list_e_r_pair_idx' in input_graph.graph_attributes.keys():
            list_e_r_pair_idx = input_graph.graph_attributes['list_e_r_pair_idx']
            list_e_e_pair_idx = None
        elif 'list_e_e_pair_idx' in input_graph.graph_attributes.keys():
            list_e_e_pair_idx = input_graph.graph_attributes['list_e_e_pair_idx']
            list_e_r_pair_idx = None
        else:
            raise RuntimeError("'list_e_r_pair_idx' or 'list_e_e_pair_idx' should be given.")

        if multi_label is None:
            input_graph.graph_attributes['logits'] = self.classifier(node_emb,
                                                                  rel_emb,
                                                                  list_e_r_pair_idx,
                                                                  list_e_e_pair_idx)
        else:
            input_graph.graph_attributes['logits'], input_graph.graph_attributes['p_score'], \
            input_graph.graph_attributes['n_score'] = self.classifier(node_emb,
                                                                   rel_emb,
                                                                   list_e_r_pair_idx,
                                                                   list_e_e_pair_idx,
                                                                   multi_label)

        return input_graph


"""
Graph Attention Networks in Grap4NLP using SPMV optimization.
Multiple heads are also batched together for faster training.
References
----------
DGL GCN example: https://github.com/dmlc/dgl/tree/master/examples/pytorch/gcn
"""
import os
import argparse
import numpy as np
import networkx as nx
import time
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.backends.cudnn as cudnn
import dgl
from dgl import DGLGraph
from dgl.data import register_data_args, load_data

from .utils import EarlyStopping
from ...modules.graph_embedding.gcn import GCN
from ...data.data import GraphData


def accuracy(logits, labels):
    _, indices = torch.max(logits, dim=1)
    correct = torch.sum(indices == labels)
    return correct.item() * 1.0 / len(labels)

def evaluate(model, g, labels, mask):
    model.eval()
    with torch.no_grad():
        logits = model(g)
        logits = logits[mask]
        labels = labels[mask]
        return accuracy(logits, labels)

class GNNClassifier(nn.Module):
    def __init__(self,
                num_layers,
                input_size,
                hidden_size,
                output_size,
                direction_option,
                norm='both',
                weight=True,
                bias=True,
                activation=F.elu,
                allow_zero_in_degree=False,
                 use_edge_weight=False):
        super(GNNClassifier, self).__init__()
        self.direction_option = direction_option
        self.model = GCN(num_layers,
                         input_size,
                         hidden_size,
                         output_size,
                         direction_option=direction_option,
                         norm=norm,
                         weight=weight,
                         bias=bias,
                         allow_zero_in_degree=allow_zero_in_degree,
                         activation=activation,
                         use_edge_weight=use_edge_weight)

        if self.direction_option == 'bi_sep':
            self.fc = nn.Linear(2 * output_size, output_size)

    def forward(self, graph):
        graph = self.model(graph)
        logits = graph.node_features['node_emb']
        if self.direction_option == 'bi_sep':
            logits = self.fc(F.elu(logits))

        return logits

def prepare_dgl_graph_data(args):
    data = load_data(args)
    features = torch.FloatTensor(data.features)
    labels = torch.LongTensor(data.labels)
    if hasattr(torch, 'BoolTensor'):
        train_mask = torch.BoolTensor(data.train_mask)
        val_mask = torch.BoolTensor(data.val_mask)
        test_mask = torch.BoolTensor(data.test_mask)
    else:
        train_mask = torch.ByteTensor(data.train_mask)
        val_mask = torch.ByteTensor(data.val_mask)
        test_mask = torch.ByteTensor(data.test_mask)

    num_feats = features.shape[1]
    n_classes = data.num_labels
    n_edges = data.graph.number_of_edges()
    print("""----Data statistics------'
      #Edges %d
      #Classes %d
      #Train samples %d
      #Val samples %d
      #Test samples %d""" %
          (n_edges, n_classes,
           train_mask.int().sum().item(),
           val_mask.int().sum().item(),
           test_mask.int().sum().item()))

    g = data.graph
    # add self loop
    g.remove_edges_from(nx.selfloop_edges(g))
    g = DGLGraph(g)
    g.add_edges(g.nodes(), g.nodes())
    n_edges = g.number_of_edges()

    data = {'features': features,
            'graph': g,
            'train_mask': train_mask,
            'val_mask': val_mask,
            'test_mask': test_mask,
            'labels': labels,
            'num_feats': num_feats,
            'n_classes': n_classes,
            'n_edges': n_edges}

    return data

def prepare_ogbn_graph_data(args):
    from ogb.nodeproppred import DglNodePropPredDataset

    dataset = DglNodePropPredDataset(name=args.dataset)

    split_idx = dataset.get_idx_split()
    train_idx, val_idx, test_idx = torch.LongTensor(split_idx['train']), torch.LongTensor(split_idx['valid']), torch.LongTensor(split_idx['test'])
    g, labels = dataset[0] # graph: dgl graph object, label: torch tensor of shape (num_nodes, num_tasks)
    features = torch.Tensor(g.ndata['feat'])
    labels = torch.LongTensor(labels).squeeze(-1)

    # add self loop
    # no duplicate self loop will be added for nodes already having self loops
    new_g = dgl.transform.add_self_loop(g)


    # edge_index = data[0]['edge_index']
    # adj = to_undirected(edge_index, num_nodes=data[0]['num_nodes'])
    # assert adj.diagonal().sum() == 0 and adj.max() <= 1 and (adj != adj.transpose()).sum() == 0

    num_feats = features.shape[1]
    n_classes = labels.max().item() + 1
    n_edges = new_g.number_of_edges()
    print("""----Data statistics------'
      #Edges %d
      #Classes %d
      #Train samples %d
      #Val samples %d
      #Test samples %d""" %
          (n_edges, n_classes,
           train_idx.shape[0],
           val_idx.shape[0],
           test_idx.shape[0]))

    data = {'features': features,
            'graph': new_g,
            'train_mask': train_idx,
            'val_mask': val_idx,
            'test_mask': test_idx,
            'labels': labels,
            'num_feats': num_feats,
            'n_classes': n_classes,
            'n_edges': n_edges}

    return data

def main(args, seed):
    # load and preprocess dataset
    if args.dataset.startswith('ogbn'):
        # Open Graph Benchmark datasets
        data = prepare_ogbn_graph_data(args)
    else:
        # DGL datasets
        data = prepare_dgl_graph_data(args)

    features, dgl_graph, train_mask, val_mask, test_mask, labels, num_feats, n_classes, n_edges\
                             = data['features'], data['graph'], data['train_mask'], \
                             data['val_mask'], data['test_mask'], data['labels'], \
                             data['num_feats'], data['n_classes'], data['n_edges']


    # Configure
    np.random.seed(seed)
    torch.manual_seed(seed)

    if not args.no_cuda and torch.cuda.is_available():
        print('[ Using CUDA ]')
        device = torch.device('cuda' if args.gpu < 0 else 'cuda:%d' % args.gpu)
        cudnn.benchmark = True
        torch.cuda.manual_seed(seed)
    else:
        device = torch.device('cpu')

    features = features.to(device)
    labels = labels.to(device)
    train_mask = train_mask.to(device)
    val_mask = val_mask.to(device)
    test_mask = test_mask.to(device)

    dgl_graph.ndata['node_feat'] = features

    # convert DGLGraph to GraphData
    g = GraphData()
    g.from_dgl(dgl_graph)

    g.edge_features['edge_weight'] = torch.tensor([1] * g.get_edge_num(), dtype=torch.float32).view(-1, 1).to(device)
    g.edge_features['reverse_edge_weight'] = torch.tensor([1] * g.get_edge_num(), dtype=torch.float32).view(-1, 1).to(
        device)

    # create model
    model = GNNClassifier(args.num_layers,
                          num_feats,
                          args.num_hidden,
                          n_classes,
                          direction_option=args.direction_option,
                          activation=F.elu,
                          use_edge_weight=False)


    print(model)
    model.to(device)

    if args.early_stop:
        stopper = EarlyStopping('{}.{}'.format(args.save_model_path, seed), patience=args.patience)

    loss_fcn = torch.nn.CrossEntropyLoss()

    # use optimizer
    optimizer = torch.optim.Adam(
        model.parameters(), lr=args.lr, weight_decay=args.weight_decay)

    # initialize graph
    dur = []
    for epoch in range(args.epochs):
        model.train()
        if epoch >= 3:
            t0 = time.time()
        # forward
        logits = model(g)
        loss = loss_fcn(logits[train_mask], labels[train_mask])

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        if epoch >= 3:
            dur.append(time.time() - t0)

        train_acc = accuracy(logits[train_mask], labels[train_mask])

        if args.fastmode:
            val_acc = accuracy(logits[val_mask], labels[val_mask])
        else:
            val_acc = evaluate(model, g, labels, val_mask)
            if args.early_stop:
                if stopper.step(val_acc, model):
                    break

        print("Epoch {:05d} | Time(s) {:.4f} | Loss {:.4f} | TrainAcc {:.4f} |"
              " ValAcc {:.4f} | ETputs(KTEPS) {:.2f}".
              format(epoch, np.mean(dur), loss.item(), train_acc,
                     val_acc, n_edges / np.mean(dur) / 1000))

    print()
    if args.early_stop:
        model = stopper.load_checkpoint(model)
        print('Restored best saved model')
        os.remove(stopper.save_model_path)
        print('Removed best saved model file to save disk space')

    acc = evaluate(model, g, labels, test_mask)
    print("Test Accuracy {:.4f}".format(acc))

    return acc


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='GCN')
    register_data_args(parser)
    parser.add_argument("--num-runs", type=int, default=5,
                        help="number of runs")
    parser.add_argument("--no-cuda", action="store_true", default=False,
                        help="use CPU")
    parser.add_argument("--gpu", type=int, default=-1,
                        help="which GPU to use.")
    parser.add_argument("--epochs", type=int, default=200,
                        help="number of training epochs")
    parser.add_argument("--direction-option", type=str, default='bi_sep',
                        help="direction type (`'undirected'`, `'bi_fuse'`, `'bi_sep'`)")
    parser.add_argument("--num-heads", type=int, default=8,
                        help="number of hidden attention heads")
    parser.add_argument("--num-out-heads", type=int, default=1,
                        help="number of output attention heads")
    parser.add_argument("--num-layers", type=int, default=2,
                        help="number of hidden layers")
    parser.add_argument("--num-hidden", type=int, default=8,
                        help="number of hidden units")
    parser.add_argument("--residual", action="store_true", default=False,
                        help="use residual connection")
    parser.add_argument("--in-drop", type=float, default=.6,
                        help="input feature dropout")
    parser.add_argument("--attn-drop", type=float, default=.6,
                        help="attention dropout")
    parser.add_argument("--lr", type=float, default=0.005,
                        help="learning rate")
    parser.add_argument('--weight-decay', type=float, default=5e-4,
                        help="weight decay")
    parser.add_argument('--negative-slope', type=float, default=0.2,
                        help="the negative slope of leaky relu")
    parser.add_argument('--early-stop', action='store_true', default=False,
                        help="indicates whether to use early stop or not")
    parser.add_argument("--patience", type=int, default=100,
                        help="early stopping patience")
    parser.add_argument('--fastmode', action="store_true", default=False,
                        help="skip re-evaluate the validation set")
    parser.add_argument('--save-model-path', type=str, default="checkpoint",
                        help="path to the best saved model")
    args = parser.parse_args()
    args.save_model_path = '{}_{}_{}_{}'.format(args.save_model_path, args.dataset, 'gcn', args.direction_option)
    print(args)

    np.random.seed(123)
    scores = []
    for _ in range(args.num_runs):
        seed = np.random.randint(10000)
        scores.append(main(args, seed))

    print("\nTest Accuracy ({} runs): mean {:.4f}, std {:.4f}".format(args.num_runs, np.mean(scores), np.std(scores)))



import torch.nn as nn


class GeneralLossBase(nn.Module):
    def __init__(self):
        super(GeneralLossBase, self).__init__()

    def forward(self, logits, label, **kwargs):
        raise NotImplementedError('GeneralLossBase: Not Implemented.')


class KGLossBase(nn.Module):
    def __init__(self):
        super(KGLossBase, self).__init__()

    def forward(self, graph, node_feat):
        raise NotImplementedError('KGLossBase: Not Implemented.')


import copy
import torch
import numpy as np

from .padding_utils import pad_2d_vals_no_size
from ...data.data import GraphData
from .vocab_utils import VocabModel, Vocab


def prepare_ext_vocab(graph_list, vocab, gt_str=None, device=None):
    """
        The function for copy mechanism.
        It will find the oov dict in ``gt_str`` and build ``oov_dict``.
        And the tgt_seq will be built by ``oov_dict``.
        Note that, the word_index in ``oov_dict`` is saved in ``token_id_oov`` of GraphData.
    Parameters
    ----------
    graph_list: list[GraphData]
        The graph list.
    vocab: VocabModel
        The VocabModel.
    gt_str: list[str], default=None
        The list contains the raw target (i.e. the ground-truth for most task) sequence.
        Note that the raw sequence should not contains <UNK>.
        It will be processed by ``oov_dict`` and generate the ``tgt_seq``.
    device: torch.device, default=None
        The device.
    Returns
    -------
    oov_dict: Vocab
        The vocab contains the out-of-vocabulary words in graph_list.
    tgt_seq: torch.Tensor
        If the ``gt_str`` is not ``None``, it will be calculated.
        We process the ``gt_str`` with ``oov_dict``.
    """
    oov_dict = copy.deepcopy(vocab.in_word_vocab)
    for g in graph_list:
        token_matrix = []
        for node_idx in range(g.get_node_num()):
            node_token = g.node_attributes[node_idx]['token']
            if oov_dict.getIndex(node_token) == oov_dict.UNK:
                oov_dict._add_words(node_token)
            token_matrix.append([oov_dict.getIndex(node_token)])
        token_matrix = torch.tensor(token_matrix, dtype=torch.long).to(device)
        g.node_features['token_id_oov'] = token_matrix

    if gt_str is not None:
        oov_tgt_collect = []
        for s in gt_str:
            oov_tgt = oov_dict.to_index_sequence(s)
            oov_tgt.append(oov_dict.EOS)
            oov_tgt = np.array(oov_tgt)
            oov_tgt_collect.append(oov_tgt)

        output_pad = pad_2d_vals_no_size(oov_tgt_collect)

        tgt_seq = torch.from_numpy(output_pad).long().to(device)
        return oov_dict, tgt_seq
    else:
        return oov_dict